# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import sfc_port_pair
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    "description": "",
    "project_id": "4ad57e7ce0b24fca8f12b9834d91079d",
    "tenant_id": "4ad57e7ce0b24fca8f12b9834d91079d",
    "egress": "d294f042-1736-11ee-821a-7f8301c71f83",
    "ingress": "d9908eba-1736-11ee-b77f-1fcc4c520068",
    "service_function_parameters": {"correlation": "mpls", "weigjt": 101},
    "id": "6ecd9cf3-ca64-46c7-863f-f2eb1b9e838a",
    "name": "port_pair_1",
}


class TestSfcPortPair(base.TestCase):
    def test_basic(self):
        sot = sfc_port_pair.SfcPortPair()
        self.assertEqual('port_pair', sot.resource_key)
        self.assertEqual('port_pairs', sot.resources_key)
        self.assertEqual('/sfc/port_pairs', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = sfc_port_pair.SfcPortPair(**EXAMPLE)
        self.assertEqual(EXAMPLE['description'], sot.description)
        self.assertEqual(EXAMPLE['egress'], sot.egress)
        self.assertEqual(EXAMPLE['ingress'], sot.ingress)
        self.assertEqual(
            EXAMPLE['service_function_parameters'],
            sot.service_function_parameters,
        )
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)

        self.assertDictEqual(
            {
                "limit": "limit",
                "marker": "marker",
                'description': 'description',
                'name': 'name',
                'project_id': 'project_id',
                'tenant_id': 'tenant_id',
                'ingress': 'ingress',
                'egress': 'egress',
            },
            sot._query_mapping._mapping,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.image.v2 import metadef_schema
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'name': 'namespace',
    'properties': {
        'namespace': {
            'type': 'string',
            'description': 'The unique namespace text.',
            'maxLength': 80,
        },
        'visibility': {
            'type': 'string',
            'description': 'Scope of namespace accessibility.',
            'enum': ['public', 'private'],
        },
        'created_at': {
            'type': 'string',
            'readOnly': True,
            'description': 'Date and time of namespace creation',
            'format': 'date-time',
        },
        'resource_type_associations': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'name': {'type': 'string'},
                    'prefix': {'type': 'string'},
                    'properties_target': {'type': 'string'},
                },
            },
        },
        'properties': {'$ref': '#/definitions/property'},
        'objects': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'name': {'type': 'string'},
                    'description': {'type': 'string'},
                    'required': {'$ref': '#/definitions/stringArray'},
                    'properties': {'$ref': '#/definitions/property'},
                },
            },
        },
        'tags': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {'name': {'type': 'string'}},
            },
        },
    },
    'additionalProperties': False,
    'definitions': {
        'positiveInteger': {'type': 'integer', 'minimum': 0},
        'positiveIntegerDefault0': {
            'allOf': [
                {'$ref': '#/definitions/positiveInteger'},
                {'default': 0},
            ]
        },
        'stringArray': {
            'type': 'array',
            'items': {'type': 'string'},
            'uniqueItems': True,
        },
    },
    'required': ['namespace'],
}


class TestMetadefSchema(base.TestCase):
    def test_basic(self):
        sot = metadef_schema.MetadefSchema()
        self.assertIsNone(sot.resource_key)
        self.assertIsNone(sot.resources_key)
        self.assertEqual('/schemas/metadefs', sot.base_path)
        self.assertFalse(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertFalse(sot.allow_list)

    def test_make_it(self):
        sot = metadef_schema.MetadefSchema(**EXAMPLE)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['properties'], sot.properties)
        self.assertEqual(
            EXAMPLE['additionalProperties'], sot.additional_properties
        )
        self.assertEqual(EXAMPLE['definitions'], sot.definitions)
        self.assertEqual(EXAMPLE['required'], sot.required)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.compute.v2 import usage
from openstack.tests.unit import base


EXAMPLE = {
    "tenant_id": "781c9299e68d4b7c80ef52712889647f",
    "server_usages": [
        {
            "hours": 79.51840531333333,
            "flavor": "m1.tiny",
            "instance_id": "76638c30-d199-4c2e-8154-7dea963bfe2f",
            "name": "test-server",
            "tenant_id": "781c9299e68d4b7c80ef52712889647f",
            "memory_mb": 512,
            "local_gb": 1,
            "vcpus": 1,
            "started_at": "2022-05-16T10:35:31.000000",
            "ended_at": None,
            "state": "active",
            "uptime": 286266,
        }
    ],
    "total_local_gb_usage": 79.51840531333333,
    "total_vcpus_usage": 79.51840531333333,
    "total_memory_mb_usage": 40713.423520426666,
    "total_hours": 79.51840531333333,
    "start": "2022-04-21T18:06:47.064959",
    "stop": "2022-05-19T18:06:37.259128",
}


class TestUsage(base.TestCase):
    def test_basic(self):
        sot = usage.Usage()
        self.assertEqual('tenant_usage', sot.resource_key)
        self.assertEqual('tenant_usages', sot.resources_key)
        self.assertEqual('/os-simple-tenant-usage', sot.base_path)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = usage.Usage(**EXAMPLE)

        self.assertEqual(EXAMPLE['tenant_id'], sot.project_id)
        self.assertEqual(
            EXAMPLE['total_local_gb_usage'],
            sot.total_local_gb_usage,
        )
        self.assertEqual(EXAMPLE['total_vcpus_usage'], sot.total_vcpus_usage)
        self.assertEqual(
            EXAMPLE['total_memory_mb_usage'],
            sot.total_memory_mb_usage,
        )
        self.assertEqual(EXAMPLE['total_hours'], sot.total_hours)
        self.assertEqual(EXAMPLE['start'], sot.start)
        self.assertEqual(EXAMPLE['stop'], sot.stop)

        # now do the embedded objects
        self.assertIsInstance(sot.server_usages, list)
        self.assertEqual(1, len(sot.server_usages))

        ssot = sot.server_usages[0]
        self.assertIsInstance(ssot, usage.ServerUsage)
        self.assertEqual(EXAMPLE['server_usages'][0]['hours'], ssot.hours)
        self.assertEqual(EXAMPLE['server_usages'][0]['flavor'], ssot.flavor)
        self.assertEqual(
            EXAMPLE['server_usages'][0]['instance_id'], ssot.instance_id
        )
        self.assertEqual(EXAMPLE['server_usages'][0]['name'], ssot.name)
        self.assertEqual(
            EXAMPLE['server_usages'][0]['tenant_id'], ssot.project_id
        )
        self.assertEqual(
            EXAMPLE['server_usages'][0]['memory_mb'], ssot.memory_mb
        )
        self.assertEqual(
            EXAMPLE['server_usages'][0]['local_gb'], ssot.local_gb
        )
        self.assertEqual(EXAMPLE['server_usages'][0]['vcpus'], ssot.vcpus)
        self.assertEqual(
            EXAMPLE['server_usages'][0]['started_at'], ssot.started_at
        )
        self.assertEqual(
            EXAMPLE['server_usages'][0]['ended_at'], ssot.ended_at
        )
        self.assertEqual(EXAMPLE['server_usages'][0]['state'], ssot.state)
        self.assertEqual(EXAMPLE['server_usages'][0]['uptime'], ssot.uptime)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
from unittest import mock

from keystoneauth1 import adapter

from openstack.baremetal.v1 import _common
from openstack.baremetal.v1 import node
from openstack import exceptions
from openstack import resource
from openstack.tests.unit import base
from openstack import utils

# NOTE: Sample data from api-ref doc
FAKE = {
    "automated_clean": False,
    "boot_mode": "uefi",
    "chassis_uuid": "1",  # NOTE: missed in api-ref sample
    "clean_step": {},
    "conductor_group": None,
    "console_enabled": False,
    "created_at": "2016-08-18T22:28:48.643434+00:00",
    "description": "A node.",
    "driver": "agent_ipmitool",
    "driver_info": {"ipmi_password": "******", "ipmi_username": "ADMIN"},
    "driver_internal_info": {},
    "extra": {},
    "firmware_interface": None,
    "inspection_finished_at": None,
    "inspection_started_at": None,
    "instance_info": {},
    "instance_name": "test-instance",
    "instance_uuid": None,
    "last_error": None,
    "lessee": None,
    "links": [
        {"href": "http://127.0.0.1:6385/v1/nodes/<NODE_ID>", "rel": "self"},
        {"href": "http://127.0.0.1:6385/nodes/<NODE_ID>", "rel": "bookmark"},
    ],
    "maintenance": False,
    "maintenance_reason": None,
    "name": "test_node",
    "network_interface": "flat",
    "owner": "4b7ed919-e4a6-4017-a081-43205c5b0b73",
    "parent_node": None,
    "portgroups": [
        {
            "href": "http://127.0.0.1:6385/v1/nodes/<NODE_ID>/portgroups",
            "rel": "self",
        },
        {
            "href": "http://127.0.0.1:6385/nodes/<NODE_ID>/portgroups",
            "rel": "bookmark",
        },
    ],
    "ports": [
        {
            "href": "http://127.0.0.1:6385/v1/nodes/<NODE_ID>/ports",
            "rel": "self",
        },
        {
            "href": "http://127.0.0.1:6385/nodes/<NODE_ID>/ports",
            "rel": "bookmark",
        },
    ],
    "power_state": None,
    "properties": {},
    "provision_state": "enroll",
    "provision_updated_at": None,
    "raid_config": {},
    "reservation": None,
    "resource_class": None,
    "service_step": {},
    "secure_boot": True,
    "shard": "TestShard",
    "runbook": None,
    "states": [
        {
            "href": "http://127.0.0.1:6385/v1/nodes/<NODE_ID>/states",
            "rel": "self",
        },
        {
            "href": "http://127.0.0.1:6385/nodes/<NODE_ID>/states",
            "rel": "bookmark",
        },
    ],
    "target_power_state": None,
    "target_provision_state": None,
    "target_raid_config": {},
    "updated_at": None,
    "uuid": "6d85703a-565d-469a-96ce-30b6de53079d",
}


def _fake_assert(self, session, expected, error_message=None):
    return expected


class TestNode(base.TestCase):
    def test_basic(self):
        sot = node.Node()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('nodes', sot.resources_key)
        self.assertEqual('/nodes', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertEqual('PATCH', sot.commit_method)

    def test_instantiate(self):
        sot = node.Node(**FAKE)

        self.assertEqual(FAKE['uuid'], sot.id)
        self.assertEqual(FAKE['name'], sot.name)
        self.assertEqual(
            FAKE['automated_clean'], sot.is_automated_clean_enabled
        )
        self.assertEqual(FAKE['boot_mode'], sot.boot_mode)
        self.assertEqual(FAKE['chassis_uuid'], sot.chassis_id)
        self.assertEqual(FAKE['clean_step'], sot.clean_step)
        self.assertEqual(FAKE['conductor_group'], sot.conductor_group)
        self.assertEqual(FAKE['created_at'], sot.created_at)
        self.assertEqual(FAKE['description'], sot.description)
        self.assertEqual(FAKE['driver'], sot.driver)
        self.assertEqual(FAKE['driver_info'], sot.driver_info)
        self.assertEqual(
            FAKE['driver_internal_info'], sot.driver_internal_info
        )
        self.assertEqual(FAKE['extra'], sot.extra)
        self.assertEqual(FAKE['firmware_interface'], sot.firmware_interface)
        self.assertEqual(FAKE['instance_info'], sot.instance_info)
        self.assertEqual(FAKE['instance_name'], sot.instance_name)
        self.assertEqual(FAKE['instance_uuid'], sot.instance_id)
        self.assertEqual(FAKE['console_enabled'], sot.is_console_enabled)
        self.assertEqual(FAKE['maintenance'], sot.is_maintenance)
        self.assertEqual(FAKE['last_error'], sot.last_error)
        self.assertEqual(FAKE['lessee'], sot.lessee)
        self.assertEqual(FAKE['links'], sot.links)
        self.assertEqual(FAKE['maintenance_reason'], sot.maintenance_reason)
        self.assertEqual(FAKE['name'], sot.name)
        self.assertEqual(FAKE['network_interface'], sot.network_interface)
        self.assertEqual(FAKE['owner'], sot.owner)
        self.assertEqual(FAKE['parent_node'], sot.parent_node)
        self.assertEqual(FAKE['ports'], sot.ports)
        self.assertEqual(FAKE['portgroups'], sot.port_groups)
        self.assertEqual(FAKE['power_state'], sot.power_state)
        self.assertEqual(FAKE['properties'], sot.properties)
        self.assertEqual(FAKE['provision_state'], sot.provision_state)
        self.assertEqual(FAKE['raid_config'], sot.raid_config)
        self.assertEqual(FAKE['reservation'], sot.reservation)
        self.assertEqual(FAKE['resource_class'], sot.resource_class)
        self.assertEqual(FAKE['service_step'], sot.service_step)
        self.assertEqual(FAKE['secure_boot'], sot.is_secure_boot)
        self.assertEqual(FAKE['runbook'], sot.runbook)
        self.assertEqual(FAKE['states'], sot.states)
        self.assertEqual(
            FAKE['target_provision_state'], sot.target_provision_state
        )
        self.assertEqual(FAKE['target_power_state'], sot.target_power_state)
        self.assertEqual(FAKE['target_raid_config'], sot.target_raid_config)
        self.assertEqual(FAKE['updated_at'], sot.updated_at)

    def test_normalize_provision_state(self):
        attrs = dict(FAKE, provision_state=None)
        sot = node.Node(**attrs)
        self.assertEqual('available', sot.provision_state)

    @mock.patch.object(node.Node, '_assert_microversion_for', _fake_assert)
    @mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
    def test_list(self):
        self.node = node.Node()
        self.session = mock.Mock(
            spec=adapter.Adapter, default_microversion=None
        )
        # Set a default, so we don't try and figure out the microversions
        # with additional requests.
        self.session.default_microversion = float(self.node._max_microversion)
        self.session.get.return_value.json.return_value = {'nodes': []}

        result = list(
            self.node.list(
                self.session,
                details=False,
                shard='meow',
                allow_unknown_params=True,
            )
        )
        self.assertEqual(0, len(result))
        self.session.get.assert_called_once_with(
            '/nodes',
            headers={'Accept': 'application/json'},
            params={'shard': 'meow'},
            microversion=float(self.node._max_microversion),
        )


@mock.patch('time.sleep', lambda _t: None)
@mock.patch.object(node.Node, 'fetch', autospec=True)
class TestNodeWaitForProvisionState(base.TestCase):
    def setUp(self):
        super().setUp()
        self.node = node.Node(**FAKE)
        self.session = mock.Mock()

    def test_success(self, mock_fetch):
        def _get_side_effect(_self, session):
            self.node.provision_state = 'manageable'
            self.assertIs(session, self.session)

        mock_fetch.side_effect = _get_side_effect

        node = self.node.wait_for_provision_state(self.session, 'manageable')
        self.assertIs(node, self.node)

    def test_failure(self, mock_fetch):
        def _get_side_effect(_self, session):
            self.node.provision_state = 'deploy failed'
            self.assertIs(session, self.session)

        mock_fetch.side_effect = _get_side_effect

        self.assertRaisesRegex(
            exceptions.ResourceFailure,
            "failure state 'deploy failed'",
            self.node.wait_for_provision_state,
            self.session,
            'manageable',
        )

    def test_failure_error(self, mock_fetch):
        def _get_side_effect(_self, session):
            self.node.provision_state = 'error'
            self.assertIs(session, self.session)

        mock_fetch.side_effect = _get_side_effect

        self.assertRaisesRegex(
            exceptions.ResourceFailure,
            "failure state 'error'",
            self.node.wait_for_provision_state,
            self.session,
            'manageable',
        )

    def test_enroll_as_failure(self, mock_fetch):
        def _get_side_effect(_self, session):
            self.node.provision_state = 'enroll'
            self.node.last_error = 'power failure'
            self.assertIs(session, self.session)

        mock_fetch.side_effect = _get_side_effect

        self.assertRaisesRegex(
            exceptions.ResourceFailure,
            'failed to verify management credentials',
            self.node.wait_for_provision_state,
            self.session,
            'manageable',
        )

    def test_timeout(self, mock_fetch):
        self.assertRaises(
            exceptions.ResourceTimeout,
            self.node.wait_for_provision_state,
            self.session,
            'manageable',
            timeout=0.001,
        )

    def test_not_abort_on_failed_state(self, mock_fetch):
        def _get_side_effect(_self, session):
            self.node.provision_state = 'deploy failed'
            self.assertIs(session, self.session)

        mock_fetch.side_effect = _get_side_effect

        self.assertRaises(
            exceptions.ResourceTimeout,
            self.node.wait_for_provision_state,
            self.session,
            'manageable',
            timeout=0.001,
            abort_on_failed_state=False,
        )


@mock.patch.object(node.Node, '_assert_microversion_for', _fake_assert)
@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
class TestNodeSetProvisionState(base.TestCase):
    def setUp(self):
        super().setUp()
        self.node = node.Node(**FAKE)
        self.session = mock.Mock(
            spec=adapter.Adapter, default_microversion=None
        )

    def test_no_arguments(self):
        result = self.node.set_provision_state(self.session, 'active')
        self.assertIs(result, self.node)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/states/provision',
            json={'target': 'active'},
            headers=mock.ANY,
            microversion=None,
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_manage(self):
        result = self.node.set_provision_state(self.session, 'manage')
        self.assertIs(result, self.node)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/states/provision',
            json={'target': 'manage'},
            headers=mock.ANY,
            microversion='1.4',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_deploy_with_configdrive(self):
        result = self.node.set_provision_state(
            self.session, 'active', config_drive='abcd'
        )
        self.assertIs(result, self.node)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/states/provision',
            json={'target': 'active', 'configdrive': 'abcd'},
            headers=mock.ANY,
            microversion=None,
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_deploy_with_configdrive_as_bytestring(self):
        config_drive = base64.b64encode(b'foo')
        result = self.node.set_provision_state(
            self.session, 'active', config_drive=config_drive
        )
        self.assertIs(result, self.node)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/states/provision',
            json={'target': 'active', 'configdrive': config_drive.decode()},
            headers=mock.ANY,
            microversion=None,
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_rebuild_with_configdrive(self):
        result = self.node.set_provision_state(
            self.session, 'rebuild', config_drive='abcd'
        )
        self.assertIs(result, self.node)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/states/provision',
            json={'target': 'rebuild', 'configdrive': 'abcd'},
            headers=mock.ANY,
            microversion='1.35',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_configdrive_as_dict(self):
        for target in ('rebuild', 'active'):
            self.session.put.reset_mock()
            result = self.node.set_provision_state(
                self.session, target, config_drive={'user_data': 'abcd'}
            )
            self.assertIs(result, self.node)
            self.session.put.assert_called_once_with(
                f'nodes/{self.node.id}/states/provision',
                json={'target': target, 'configdrive': {'user_data': 'abcd'}},
                headers=mock.ANY,
                microversion='1.56',
                retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
            )

    def test_deploy_with_deploy_steps(self):
        deploy_steps = [{'interface': 'deploy', 'step': 'upgrade_fw'}]
        result = self.node.set_provision_state(
            self.session, 'active', deploy_steps=deploy_steps
        )

        self.assertIs(result, self.node)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/states/provision',
            json={'target': 'active', 'deploy_steps': deploy_steps},
            headers=mock.ANY,
            microversion='1.69',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_rebuild_with_deploy_steps(self):
        deploy_steps = [{'interface': 'deploy', 'step': 'upgrade_fw'}]
        result = self.node.set_provision_state(
            self.session, 'rebuild', deploy_steps=deploy_steps
        )

        self.assertIs(result, self.node)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/states/provision',
            json={'target': 'rebuild', 'deploy_steps': deploy_steps},
            headers=mock.ANY,
            microversion='1.69',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_set_provision_state_unhold(self):
        result = self.node.set_provision_state(self.session, 'unhold')

        self.assertIs(result, self.node)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/states/provision',
            json={'target': 'unhold'},
            headers=mock.ANY,
            microversion='1.85',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_set_provision_state_service(self):
        service_steps = [{'interface': 'deploy', 'step': 'hold'}]
        result = self.node.set_provision_state(
            self.session, 'service', service_steps=service_steps
        )

        self.assertIs(result, self.node)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/states/provision',
            json={'target': 'service', 'service_steps': service_steps},
            headers=mock.ANY,
            microversion='1.87',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_set_provision_state_clean_runbook(self):
        runbook = 'CUSTOM_AWESOME'
        result = self.node.set_provision_state(
            self.session, 'clean', runbook=runbook
        )

        self.assertIs(result, self.node)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/states/provision',
            json={'target': 'clean', 'runbook': runbook},
            headers=mock.ANY,
            microversion='1.92',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_set_provision_state_service_runbook(self):
        runbook = 'CUSTOM_AWESOME'
        result = self.node.set_provision_state(
            self.session, 'service', runbook=runbook
        )

        self.assertIs(result, self.node)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/states/provision',
            json={'target': 'service', 'runbook': runbook},
            headers=mock.ANY,
            microversion='1.92',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )


@mock.patch.object(node.Node, '_translate_response', mock.Mock())
@mock.patch.object(node.Node, '_get_session', lambda self, x: x)
@mock.patch.object(node.Node, 'set_provision_state', autospec=True)
class TestNodeCreate(base.TestCase):
    def setUp(self):
        super().setUp()
        self.new_state = None
        self.session = mock.Mock(spec=adapter.Adapter)
        self.session.default_microversion = '1.1'
        self.node = node.Node(driver=FAKE['driver'])

        def _change_state(*args, **kwargs):
            self.node.provision_state = self.new_state

        self.session.post.side_effect = _change_state

    def test_available_old_version(self, mock_prov):
        self.node.provision_state = 'available'
        result = self.node.create(self.session)
        self.assertIs(result, self.node)
        self.session.post.assert_called_once_with(
            mock.ANY,
            json={'driver': FAKE['driver']},
            headers=mock.ANY,
            microversion=self.session.default_microversion,
            params={},
        )
        self.assertFalse(mock_prov.called)

    def test_available_new_version(self, mock_prov):
        self.session.default_microversion = '1.11'
        self.node.provision_state = 'available'

        result = self.node.create(self.session)
        self.assertIs(result, self.node)
        self.session.post.assert_called_once_with(
            mock.ANY,
            json={'driver': FAKE['driver']},
            headers=mock.ANY,
            microversion='1.10',
            params={},
        )
        mock_prov.assert_not_called()

    def test_no_enroll_in_old_version(self, mock_prov):
        self.node.provision_state = 'enroll'
        self.assertRaises(
            exceptions.NotSupported, self.node.create, self.session
        )
        self.assertFalse(self.session.post.called)
        self.assertFalse(mock_prov.called)

    def test_enroll_new_version(self, mock_prov):
        self.session.default_microversion = '1.11'
        self.node.provision_state = 'enroll'
        self.new_state = 'enroll'

        result = self.node.create(self.session)
        self.assertIs(result, self.node)
        self.session.post.assert_called_once_with(
            mock.ANY,
            json={'driver': FAKE['driver']},
            headers=mock.ANY,
            microversion=self.session.default_microversion,
            params={},
        )
        self.assertFalse(mock_prov.called)

    def test_no_manageable_in_old_version(self, mock_prov):
        self.node.provision_state = 'manageable'
        self.assertRaises(
            exceptions.NotSupported, self.node.create, self.session
        )
        self.assertFalse(self.session.post.called)
        self.assertFalse(mock_prov.called)

    def test_manageable_old_version(self, mock_prov):
        self.session.default_microversion = '1.4'
        self.node.provision_state = 'manageable'
        self.new_state = 'available'

        result = self.node.create(self.session)
        self.assertIs(result, self.node)
        self.session.post.assert_called_once_with(
            mock.ANY,
            json={'driver': FAKE['driver']},
            headers=mock.ANY,
            microversion=self.session.default_microversion,
            params={},
        )
        mock_prov.assert_called_once_with(
            self.node, self.session, 'manage', wait=True
        )

    def test_manageable_new_version(self, mock_prov):
        self.session.default_microversion = '1.11'
        self.node.provision_state = 'manageable'
        self.new_state = 'enroll'

        result = self.node.create(self.session)
        self.assertIs(result, self.node)
        self.session.post.assert_called_once_with(
            mock.ANY,
            json={'driver': FAKE['driver']},
            headers=mock.ANY,
            microversion=self.session.default_microversion,
            params={},
        )
        mock_prov.assert_called_once_with(
            self.node, self.session, 'manage', wait=True
        )


@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
@mock.patch.object(node.Node, '_get_session', lambda self, x: x)
class TestNodeVif(base.TestCase):
    def setUp(self):
        super().setUp()
        self.session = mock.Mock(spec=adapter.Adapter)
        self.session.default_microversion = '1.67'
        self.session.log = mock.Mock()
        self.node = node.Node(
            id='c29db401-b6a7-4530-af8e-20a720dee946', driver=FAKE['driver']
        )
        self.vif_id = '714bdf6d-2386-4b5e-bd0d-bc036f04b1ef'
        self.vif_port_uuid = 'port-uuid'
        self.vif_portgroup_uuid = 'portgroup-uuid'

    def test_attach_vif(self):
        self.assertIsNone(self.node.attach_vif(self.session, self.vif_id))
        self.session.post.assert_called_once_with(
            f'nodes/{self.node.id}/vifs',
            json={'id': self.vif_id},
            headers=mock.ANY,
            microversion='1.67',
            retriable_status_codes=[409, 503],
        )

    def test_attach_vif_no_retries(self):
        self.assertIsNone(
            self.node.attach_vif(
                self.session, self.vif_id, retry_on_conflict=False
            )
        )
        self.session.post.assert_called_once_with(
            f'nodes/{self.node.id}/vifs',
            json={'id': self.vif_id},
            headers=mock.ANY,
            microversion='1.67',
            retriable_status_codes=[503],
        )

    def test_attach_vif_with_port_uuid(self):
        self.assertIsNone(
            self.node.attach_vif(
                self.session, self.vif_id, port_id=self.vif_port_uuid
            )
        )
        self.session.post.assert_called_once_with(
            f'nodes/{self.node.id}/vifs',
            json={'id': self.vif_id, 'port_uuid': self.vif_port_uuid},
            headers=mock.ANY,
            microversion='1.67',
            retriable_status_codes=[409, 503],
        )

    def test_attach_vif_with_portgroup_uuid(self):
        self.assertIsNone(
            self.node.attach_vif(
                self.session,
                self.vif_id,
                port_group_id=self.vif_portgroup_uuid,
            )
        )
        self.session.post.assert_called_once_with(
            f'nodes/{self.node.id}/vifs',
            json={
                'id': self.vif_id,
                'portgroup_uuid': self.vif_portgroup_uuid,
            },
            headers=mock.ANY,
            microversion='1.67',
            retriable_status_codes=[409, 503],
        )

    def test_attach_vif_with_port_uuid_and_portgroup_uuid(self):
        self.assertRaises(
            exceptions.InvalidRequest,
            self.node.attach_vif,
            self.session,
            self.vif_id,
            port_id=self.vif_port_uuid,
            port_group_id=self.vif_portgroup_uuid,
        )

    def test_detach_vif_existing(self):
        self.assertTrue(self.node.detach_vif(self.session, self.vif_id))
        self.session.delete.assert_called_once_with(
            f'nodes/{self.node.id}/vifs/{self.vif_id}',
            headers=mock.ANY,
            microversion='1.67',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_detach_vif_missing(self):
        self.session.delete.return_value.status_code = 400
        self.assertFalse(self.node.detach_vif(self.session, self.vif_id))
        self.session.delete.assert_called_once_with(
            f'nodes/{self.node.id}/vifs/{self.vif_id}',
            headers=mock.ANY,
            microversion='1.67',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_list_vifs(self):
        self.session.get.return_value.json.return_value = {
            'vifs': [
                {'id': '1234'},
                {'id': '5678'},
            ]
        }
        res = self.node.list_vifs(self.session)
        self.assertEqual(['1234', '5678'], res)
        self.session.get.assert_called_once_with(
            f'nodes/{self.node.id}/vifs',
            headers=mock.ANY,
            microversion='1.67',
        )

    def test_incompatible_microversion(self):
        self.session.default_microversion = '1.1'
        self.assertRaises(
            exceptions.NotSupported,
            self.node.attach_vif,
            self.session,
            self.vif_id,
        )
        self.assertRaises(
            exceptions.NotSupported,
            self.node.detach_vif,
            self.session,
            self.vif_id,
        )
        self.assertRaises(
            exceptions.NotSupported, self.node.list_vifs, self.session
        )

    def test_incompatible_microversion_optional_params(self):
        self.session.default_microversion = '1.28'
        self.assertRaises(
            exceptions.NotSupported,
            self.node.attach_vif,
            self.session,
            self.vif_id,
            port_id=self.vif_port_uuid,
        )
        self.assertRaises(
            exceptions.NotSupported,
            self.node.attach_vif,
            self.session,
            self.vif_id,
            port_group_id=self.vif_portgroup_uuid,
        )


@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
@mock.patch.object(node.Node, '_get_session', lambda self, x: x)
class TestNodeVmedia(base.TestCase):
    def setUp(self):
        super().setUp()
        self.session = mock.Mock(spec=adapter.Adapter)
        self.session.default_microversion = '1.89'
        self.session.log = mock.Mock()
        self.node = node.Node(
            id='c29db401-b6a7-4530-af8e-20a720dee946', driver=FAKE['driver']
        )
        self.device_type = "CDROM"
        self.image_url = "http://image"

    def test_attach_vmedia(self):
        self.assertIsNone(
            self.node.attach_vmedia(
                self.session, self.device_type, self.image_url
            )
        )
        self.session.post.assert_called_once_with(
            f'nodes/{self.node.id}/vmedia',
            json={
                'device_type': self.device_type,
                'image_url': self.image_url,
            },
            headers=mock.ANY,
            microversion='1.89',
            retriable_status_codes=[409, 503],
        )

    def test_attach_vmedia_no_retries(self):
        self.assertIsNone(
            self.node.attach_vmedia(
                self.session,
                self.device_type,
                self.image_url,
                retry_on_conflict=False,
            )
        )
        self.session.post.assert_called_once_with(
            f'nodes/{self.node.id}/vmedia',
            json={
                'device_type': self.device_type,
                'image_url': self.image_url,
            },
            headers=mock.ANY,
            microversion='1.89',
            retriable_status_codes=[503],
        )

    def test_detach_vmedia_existing(self):
        self.assertIsNone(self.node.detach_vmedia(self.session))
        self.session.delete.assert_called_once_with(
            f'nodes/{self.node.id}/vmedia',
            headers=mock.ANY,
            microversion='1.89',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_detach_vmedia_missing(self):
        self.session.delete.return_value.status_code = 400
        self.assertIsNone(self.node.detach_vmedia(self.session))
        self.session.delete.assert_called_once_with(
            f'nodes/{self.node.id}/vmedia',
            headers=mock.ANY,
            microversion='1.89',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_incompatible_microversion(self):
        self.session.default_microversion = '1.1'
        self.assertRaises(
            exceptions.NotSupported,
            self.node.attach_vmedia,
            self.session,
            self.device_type,
            self.image_url,
        )
        self.assertRaises(
            exceptions.NotSupported,
            self.node.detach_vmedia,
            self.session,
        )


@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
@mock.patch.object(node.Node, '_get_session', lambda self, x: x)
class TestNodeValidate(base.TestCase):
    def setUp(self):
        super().setUp()
        self.session = mock.Mock(spec=adapter.Adapter)
        self.session.default_microversion = '1.28'
        self.node = node.Node(**FAKE)

    def test_validate_ok(self):
        self.session.get.return_value.json.return_value = {
            'boot': {'result': True},
            'console': {'result': False, 'reason': 'Not configured'},
            'deploy': {'result': True},
            'inspect': {'result': None, 'reason': 'Not supported'},
            'power': {'result': True},
        }
        result = self.node.validate(self.session)
        for iface in ('boot', 'deploy', 'power'):
            self.assertTrue(result[iface].result)
            self.assertFalse(result[iface].reason)
        for iface in ('console', 'inspect'):
            self.assertIsNot(True, result[iface].result)
            self.assertTrue(result[iface].reason)

    def test_validate_failed(self):
        self.session.get.return_value.json.return_value = {
            'boot': {'result': False},
            'console': {'result': False, 'reason': 'Not configured'},
            'deploy': {'result': False, 'reason': 'No deploy for you'},
            'inspect': {'result': None, 'reason': 'Not supported'},
            'power': {'result': True},
        }
        self.assertRaisesRegex(
            exceptions.ValidationException,
            'No deploy for you',
            self.node.validate,
            self.session,
        )

    def test_validate_no_failure(self):
        self.session.get.return_value.json.return_value = {
            'boot': {'result': False},
            'console': {'result': False, 'reason': 'Not configured'},
            'deploy': {'result': False, 'reason': 'No deploy for you'},
            'inspect': {'result': None, 'reason': 'Not supported'},
            'power': {'result': True},
        }
        result = self.node.validate(self.session, required=None)
        self.assertTrue(result['power'].result)
        self.assertFalse(result['power'].reason)
        for iface in ('deploy', 'console', 'inspect'):
            self.assertIsNot(True, result[iface].result)
            self.assertTrue(result[iface].reason)
        # Reason can be empty
        self.assertFalse(result['boot'].result)
        self.assertIsNone(result['boot'].reason)


@mock.patch('time.sleep', lambda _t: None)
@mock.patch.object(node.Node, 'fetch', autospec=True)
class TestNodeWaitForReservation(base.TestCase):
    def setUp(self):
        super().setUp()
        self.session = mock.Mock(spec=adapter.Adapter)
        self.session.default_microversion = '1.6'
        self.session.log = mock.Mock()
        self.node = node.Node(**FAKE)

    def test_no_reservation(self, mock_fetch):
        self.node.reservation = None
        node = self.node.wait_for_reservation(None)
        self.assertIs(node, self.node)
        self.assertFalse(mock_fetch.called)

    def test_reservation(self, mock_fetch):
        self.node.reservation = 'example.com'

        def _side_effect(node, session):
            if self.node.reservation == 'example.com':
                self.node.reservation = 'example2.com'
            else:
                self.node.reservation = None

        mock_fetch.side_effect = _side_effect
        node = self.node.wait_for_reservation(self.session)
        self.assertIs(node, self.node)
        self.assertEqual(2, mock_fetch.call_count)

    def test_timeout(self, mock_fetch):
        self.node.reservation = 'example.com'

        self.assertRaises(
            exceptions.ResourceTimeout,
            self.node.wait_for_reservation,
            self.session,
            timeout=0.001,
        )
        mock_fetch.assert_called_with(self.node, self.session)


@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
class TestNodeInjectNMI(base.TestCase):
    def setUp(self):
        super().setUp()
        self.node = node.Node(**FAKE)
        self.session = mock.Mock(spec=adapter.Adapter)
        self.session.default_microversion = '1.29'
        self.node = node.Node(**FAKE)

    def test_inject_nmi(self):
        self.node.inject_nmi(self.session)
        self.session.put.assert_called_once_with(
            'nodes/{}/management/inject_nmi'.format(FAKE['uuid']),
            json={},
            headers=mock.ANY,
            microversion='1.29',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_incompatible_microversion(self):
        self.session.default_microversion = '1.28'
        self.assertRaises(
            exceptions.NotSupported,
            self.node.inject_nmi,
            self.session,
        )


@mock.patch.object(node.Node, '_assert_microversion_for', _fake_assert)
@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
class TestNodeSetPowerState(base.TestCase):
    def setUp(self):
        super().setUp()
        self.node = node.Node(**FAKE)
        self.session = mock.Mock(
            spec=adapter.Adapter, default_microversion=None
        )

    def test_power_on(self):
        self.node.set_power_state(self.session, 'power on')
        self.session.put.assert_called_once_with(
            'nodes/{}/states/power'.format(FAKE['uuid']),
            json={'target': 'power on'},
            headers=mock.ANY,
            microversion=None,
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_soft_power_on(self):
        self.node.set_power_state(self.session, 'soft power off')
        self.session.put.assert_called_once_with(
            'nodes/{}/states/power'.format(FAKE['uuid']),
            json={'target': 'soft power off'},
            headers=mock.ANY,
            microversion='1.27',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )


@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
@mock.patch.object(node.Node, '_translate_response', mock.Mock())
@mock.patch.object(node.Node, '_get_session', lambda self, x: x)
class TestNodeMaintenance(base.TestCase):
    def setUp(self):
        super().setUp()
        self.node = node.Node.existing(**FAKE)
        self.session = mock.Mock(
            spec=adapter.Adapter,
            default_microversion='1.1',
            retriable_status_codes=None,
        )

    def test_set(self):
        self.node.set_maintenance(self.session)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/maintenance',
            json={'reason': None},
            headers=mock.ANY,
            microversion=mock.ANY,
        )

    def test_set_with_reason(self):
        self.node.set_maintenance(self.session, 'No work on Monday')
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/maintenance',
            json={'reason': 'No work on Monday'},
            headers=mock.ANY,
            microversion=mock.ANY,
        )

    def test_unset(self):
        self.node.unset_maintenance(self.session)
        self.session.delete.assert_called_once_with(
            f'nodes/{self.node.id}/maintenance',
            json=None,
            headers=mock.ANY,
            microversion=mock.ANY,
        )

    def test_set_via_update(self):
        self.node.is_maintenance = True
        self.node.commit(self.session)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/maintenance',
            json={'reason': None},
            headers=mock.ANY,
            microversion=mock.ANY,
        )

        self.assertFalse(self.session.patch.called)

    def test_set_with_reason_via_update(self):
        self.node.is_maintenance = True
        self.node.maintenance_reason = 'No work on Monday'
        self.node.commit(self.session)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/maintenance',
            json={'reason': 'No work on Monday'},
            headers=mock.ANY,
            microversion=mock.ANY,
        )
        self.assertFalse(self.session.patch.called)

    def test_set_with_other_fields(self):
        self.node.is_maintenance = True
        self.node.name = 'lazy-3000'
        self.node.commit(self.session)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/maintenance',
            json={'reason': None},
            headers=mock.ANY,
            microversion=mock.ANY,
        )

        self.session.patch.assert_called_once_with(
            f'nodes/{self.node.id}',
            json=[{'path': '/name', 'op': 'replace', 'value': 'lazy-3000'}],
            headers=mock.ANY,
            microversion=mock.ANY,
        )

    def test_set_with_reason_and_other_fields(self):
        self.node.is_maintenance = True
        self.node.maintenance_reason = 'No work on Monday'
        self.node.name = 'lazy-3000'
        self.node.commit(self.session)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/maintenance',
            json={'reason': 'No work on Monday'},
            headers=mock.ANY,
            microversion=mock.ANY,
        )

        self.session.patch.assert_called_once_with(
            f'nodes/{self.node.id}',
            json=[{'path': '/name', 'op': 'replace', 'value': 'lazy-3000'}],
            headers=mock.ANY,
            microversion=mock.ANY,
        )

    def test_no_reason_without_maintenance(self):
        self.node.maintenance_reason = 'Can I?'
        self.assertRaises(ValueError, self.node.commit, self.session)
        self.assertFalse(self.session.put.called)
        self.assertFalse(self.session.patch.called)

    def test_set_unset_maintenance(self):
        self.node.is_maintenance = True
        self.node.maintenance_reason = 'No work on Monday'
        self.node.commit(self.session)

        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/maintenance',
            json={'reason': 'No work on Monday'},
            headers=mock.ANY,
            microversion=mock.ANY,
        )

        self.node.is_maintenance = False
        self.node.commit(self.session)
        self.assertIsNone(self.node.maintenance_reason)

        self.session.delete.assert_called_once_with(
            f'nodes/{self.node.id}/maintenance',
            json=None,
            headers=mock.ANY,
            microversion=mock.ANY,
        )


@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
class TestNodeBootDevice(base.TestCase):
    def setUp(self):
        super().setUp()
        self.node = node.Node(**FAKE)
        self.session = mock.Mock(
            spec=adapter.Adapter, default_microversion='1.1'
        )

    def test_get_boot_device(self):
        self.node.get_boot_device(self.session)
        self.session.get.assert_called_once_with(
            f'nodes/{self.node.id}/management/boot_device',
            headers=mock.ANY,
            microversion=mock.ANY,
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_set_boot_device(self):
        self.node.set_boot_device(self.session, 'pxe', persistent=False)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/management/boot_device',
            json={'boot_device': 'pxe', 'persistent': False},
            headers=mock.ANY,
            microversion=mock.ANY,
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_get_supported_boot_devices(self):
        self.node.get_supported_boot_devices(self.session)
        self.session.get.assert_called_once_with(
            f'nodes/{self.node.id}/management/boot_device/supported',
            headers=mock.ANY,
            microversion=mock.ANY,
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )


@mock.patch.object(utils, 'pick_microversion', lambda session, v: v)
@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
class TestNodeSetBootMode(base.TestCase):
    def setUp(self):
        super().setUp()
        self.node = node.Node(**FAKE)
        self.session = mock.Mock(
            spec=adapter.Adapter, default_microversion='1.1'
        )

    def test_node_set_boot_mode(self):
        self.node.set_boot_mode(self.session, 'uefi')
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/states/boot_mode',
            json={'target': 'uefi'},
            headers=mock.ANY,
            microversion=mock.ANY,
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_node_set_boot_mode_invalid_mode(self):
        self.assertRaises(
            ValueError, self.node.set_boot_mode, self.session, 'invalid-efi'
        )


@mock.patch.object(utils, 'pick_microversion', lambda session, v: v)
@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
class TestNodeSetSecureBoot(base.TestCase):
    def setUp(self):
        super().setUp()
        self.node = node.Node(**FAKE)
        self.session = mock.Mock(
            spec=adapter.Adapter, default_microversion='1.1'
        )

    def test_node_set_secure_boot(self):
        self.node.set_secure_boot(self.session, True)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/states/secure_boot',
            json={'target': True},
            headers=mock.ANY,
            microversion=mock.ANY,
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_node_set_secure_boot_invalid_none(self):
        self.assertRaises(
            ValueError, self.node.set_secure_boot, self.session, None
        )


@mock.patch.object(utils, 'pick_microversion', lambda session, v: v)
@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
class TestNodeTraits(base.TestCase):
    def setUp(self):
        super().setUp()
        self.node = node.Node(**FAKE)
        self.session = mock.Mock(
            spec=adapter.Adapter, default_microversion='1.37'
        )
        self.session.log = mock.Mock()

    def test_node_add_trait(self):
        self.node.add_trait(self.session, 'CUSTOM_FAKE')
        self.session.put.assert_called_once_with(
            'nodes/{}/traits/{}'.format(self.node.id, 'CUSTOM_FAKE'),
            json=None,
            headers=mock.ANY,
            microversion='1.37',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_remove_trait(self):
        self.assertTrue(self.node.remove_trait(self.session, 'CUSTOM_FAKE'))
        self.session.delete.assert_called_once_with(
            'nodes/{}/traits/{}'.format(self.node.id, 'CUSTOM_FAKE'),
            headers=mock.ANY,
            microversion='1.37',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_remove_trait_missing(self):
        self.session.delete.return_value.status_code = 400
        self.assertFalse(
            self.node.remove_trait(self.session, 'CUSTOM_MISSING')
        )
        self.session.delete.assert_called_once_with(
            'nodes/{}/traits/{}'.format(self.node.id, 'CUSTOM_MISSING'),
            headers=mock.ANY,
            microversion='1.37',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_set_traits(self):
        traits = ['CUSTOM_FAKE', 'CUSTOM_REAL', 'CUSTOM_MISSING']
        self.node.set_traits(self.session, traits)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/traits',
            json={'traits': ['CUSTOM_FAKE', 'CUSTOM_REAL', 'CUSTOM_MISSING']},
            headers=mock.ANY,
            microversion='1.37',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )


@mock.patch.object(node.Node, '_assert_microversion_for', _fake_assert)
@mock.patch.object(resource.Resource, 'patch', autospec=True)
class TestNodePatch(base.TestCase):
    def setUp(self):
        super().setUp()
        self.node = node.Node(**FAKE)
        self.session = mock.Mock(
            spec=adapter.Adapter,
            default_microversion='1.1',
            retriable_status_codes=None,
        )
        self.session.log = mock.Mock()

    def test_node_patch(self, mock_patch):
        patch = {'path': 'test'}
        self.node.patch(self.session, patch=patch)
        mock_patch.assert_called_once()
        kwargs = mock_patch.call_args[1]
        self.assertEqual(kwargs['patch'], {'path': 'test'})

    @mock.patch.object(resource.Resource, '_prepare_request', autospec=True)
    @mock.patch.object(resource.Resource, '_commit', autospec=True)
    def test_node_patch_reset_interfaces(
        self, mock__commit, mock_prepreq, mock_patch
    ):
        patch = {'path': 'test'}
        self.node.patch(
            self.session,
            patch=patch,
            retry_on_conflict=True,
            reset_interfaces=True,
        )
        mock_prepreq.assert_called_once()
        prepreq_kwargs = mock_prepreq.call_args[1]
        self.assertEqual(prepreq_kwargs['params'], {'reset_interfaces': True})
        mock__commit.assert_called_once()
        commit_args = mock__commit.call_args[0]
        commit_kwargs = mock__commit.call_args[1]
        self.assertIn('1.45', commit_args)
        self.assertEqual(commit_kwargs['retry_on_conflict'], True)
        mock_patch.assert_not_called()


@mock.patch('time.sleep', lambda _t: None)
@mock.patch.object(node.Node, 'fetch', autospec=True)
class TestNodeWaitForPowerState(base.TestCase):
    def setUp(self):
        super().setUp()
        self.node = node.Node(**FAKE)
        self.session = mock.Mock()

    def test_success(self, mock_fetch):
        self.node.power_state = 'power on'

        def _get_side_effect(_self, session):
            self.node.power_state = 'power off'
            self.assertIs(session, self.session)

        mock_fetch.side_effect = _get_side_effect

        node = self.node.wait_for_power_state(self.session, 'power off')
        self.assertIs(node, self.node)

    def test_timeout(self, mock_fetch):
        self.node.power_state = 'power on'
        self.assertRaises(
            exceptions.ResourceTimeout,
            self.node.wait_for_power_state,
            self.session,
            'power off',
            timeout=0.001,
        )


@mock.patch.object(utils, 'pick_microversion', lambda session, v: v)
@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
class TestNodePassthru:
    def setUp(self):
        super().setUp()
        self.node = node.Node(**FAKE)
        self.session = node.Mock(
            spec=adapter.Adapter, default_microversion='1.37'
        )
        self.session.log = mock.Mock()

    def test_get_passthru(self):
        self.node.call_vendor_passthru(self.session, "GET", "test_method")
        self.session.get.assert_called_once_with(
            f'nodes/{self.node.id}/vendor_passthru?method=test_method',
            headers=mock.ANY,
            microversion='1.37',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_post_passthru(self):
        self.node.call_vendor_passthru(self.session, "POST", "test_method")
        self.session.post.assert_called_once_with(
            f'nodes/{self.node.id}/vendor_passthru?method=test_method',
            headers=mock.ANY,
            microversion='1.37',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_put_passthru(self):
        self.node.call_vendor_passthru(self.session, "PUT", "test_method")
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/vendor_passthru?method=test_method',
            headers=mock.ANY,
            microversion='1.37',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_delete_passthru(self):
        self.node.call_vendor_passthru(self.session, "DELETE", "test_method")
        self.session.delete.assert_called_once_with(
            f'nodes/{self.node.id}/vendor_passthru?method=test_method',
            headers=mock.ANY,
            microversion='1.37',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_list_passthru(self):
        self.node.list_vendor_passthru(self.session)
        self.session.get.assert_called_once_with(
            f'nodes/{self.node.id}/vendor_passthru/methods',
            headers=mock.ANY,
            microversion='1.37',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )


@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
class TestNodeConsole(base.TestCase):
    def setUp(self):
        super().setUp()
        self.node = node.Node(**FAKE)
        self.session = mock.Mock(
            spec=adapter.Adapter,
            default_microversion='1.1',
        )

    def test_get_console(self):
        self.node.get_console(self.session)
        self.session.get.assert_called_once_with(
            f'nodes/{self.node.id}/states/console',
            headers=mock.ANY,
            microversion=mock.ANY,
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_set_console_mode(self):
        self.node.set_console_mode(self.session, True)
        self.session.put.assert_called_once_with(
            f'nodes/{self.node.id}/states/console',
            json={'enabled': True},
            headers=mock.ANY,
            microversion=mock.ANY,
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )

    def test_set_console_mode_invalid_enabled(self):
        self.assertRaises(
            ValueError,
            self.node.set_console_mode,
            self.session,
            'true',  # not a bool
        )


@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
class TestNodeInventory(base.TestCase):
    def setUp(self):
        super().setUp()
        self.node = node.Node(**FAKE)
        self.session = mock.Mock(
            spec=adapter.Adapter,
            default_microversion='1.81',
        )

    def test_get_inventory(self):
        node_inventory = {
            'inventory': {
                'memory': {'physical_mb': 3072},
                'cpu': {
                    'count': 1,
                    'model_name': 'qemu64',
                    'architecture': 'x86_64',
                },
                'disks': [{'name': 'testvm1.qcow2', 'size': 11811160064}],
                'interfaces': [{'mac_address': '52:54:00:c7:02:45'}],
                'system_vendor': {
                    'product_name': 'testvm1',
                    'manufacturer': 'Sushy Emulator',
                },
                'boot': {'current_boot_mode': 'uefi'},
            },
            'plugin_data': {'fake_plugin_data'},
        }
        self.session.get.return_value.json.return_value = node_inventory

        res = self.node.get_node_inventory(self.session, self.node.id)
        self.assertEqual(node_inventory, res)

        self.session.get.assert_called_once_with(
            f'nodes/{self.node.id}/inventory',
            headers=mock.ANY,
            microversion='1.81',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )


@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
class TestNodeFirmware(base.TestCase):
    def setUp(self):
        super().setUp()
        self.node = node.Node(**FAKE)
        self.session = mock.Mock(
            spec=adapter.Adapter,
            default_microversion='1.86',
        )

    def test_list_firmware(self):
        node_firmware = {
            "firmware": [
                {
                    "created_at": "2016-08-18T22:28:49.653974+00:00",
                    "updated_at": "2016-08-18T22:28:49.653974+00:00",
                    "component": "BMC",
                    "initial_version": "v1.0.0",
                    "current_version": "v1.2.0",
                    "last_version_flashed": "v1.2.0",
                },
                {
                    "created_at": "2016-08-18T22:28:49.653974+00:00",
                    "updated_at": "2016-08-18T22:28:49.653974+00:00",
                    "component": "BIOS",
                    "initial_version": "v1.0.0",
                    "current_version": "v1.1.5",
                    "last_version_flashed": "v1.1.5",
                },
            ]
        }
        self.session.get.return_value.json.return_value = node_firmware

        res = self.node.list_firmware(self.session)
        self.assertEqual(node_firmware, res)

        self.session.get.assert_called_once_with(
            f'nodes/{self.node.id}/firmware',
            headers=mock.ANY,
            microversion='1.86',
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Self

from keystoneauth1 import adapter

from openstack import resource, exceptions, utils


class ShareTransfer(resource.Resource):
    resource_key = "transfer"
    base_path = "/share-transfers"

    # capabilities
    allow_list = True
    allow_create = True
    allow_fetch = True
    allow_commit = False
    allow_delete = True
    allow_head = False

    _query_mapping = resource.QueryParameters(
        "id",
        "share_id",
        "name",
        "auth_key",
        "clear_access_rules",
        "all_tenants",
        "destination_project_id",
        "limit",
        "offset",
        "sort_key",
        "sort_dir",
    )
    # Properties
    #: The transfer UUID.
    id = resource.Body("id", type=str)
    #: The share UUID.
    share_id = resource.Body("share_id", type=str)
    #: The date and time stamp when the resource was created
    #: within the service's database.
    created_at = resource.Body("created_at", type=str)
    #: The transfer display name.
    name = resource.Body("name", type=str)
    #: The type of the resource for the transfer.
    resource_type = resource.Body("resource_type", type=str)
    #: The UUID of the resource for the transfer.
    resource_id = resource.Body("resource_id", type=str)
    #: The authentication key for the transfer.
    auth_key = resource.Body("auth_key", type=str)
    #: The ID of the project that owns the resource.
    source_project_id = resource.Body("source_project_id", type=str)
    #: UUID of the destination project to accept transfer resource.
    destination_project_id = resource.Body("destination_project_id", type=str)
    #: Whether the transfer has been accepted.
    accepted = resource.Body("accepted", type=bool)
    #: The date and time stamp when the resource transfer will expire
    #: After transfer expired, will be automatically deleted
    expires_at = resource.Body("expires_at", type=str)
    #: Pagination and bookmark links for the resource.
    links = resource.Body("links", type=list)

    def accept(
        self,
        session: adapter.Adapter,
        *,
        auth_key: str,
        clear_access_rules: bool = False,
    ) -> Self:
        """Accept a share transfer using the auth_key."""

        url = utils.urljoin(self.base_path, self.id, "accept")
        body = {
            "accept": {
                "auth_key": auth_key,
                "clear_access_rules": clear_access_rules,
            }
        }

        response = session.post(url, json=body)

        exceptions.raise_from_response(response)
        self._translate_response(response)
        return self



# Copyright 2026 Red Hat Inc.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from openstack import resource


class ImageLocation(resource.Resource):
    base_path = '/images/%(image_id)s/locations'

    allow_create = True
    allow_list = True

    # Note: Glance does not support microversions. This attribute is used
    # by the SDK framework to indicate the latest API capability level
    # required for this feature. The location import API requires Glance
    # API v2 with feature capability detection indicating support for
    # location management (typically available in API v2.17+).
    _max_microversion = '2.17'

    #: Image ID stored through the image API. Typically a UUID.
    image_id = resource.URI('image_id')
    #: The Location URL to access the image from external store.
    url = resource.Body('url')
    #: Optional checksum/hash fields for location validation (Glance API).
    validation_data = resource.Body('validation_data', type=dict)
    #: The location metadata.
    metadata = resource.Body('metadata', type=dict)



# Copyright 2017 OVH SAS
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_qos_bandwidth_limit_rule
----------------------------------

Functional tests for QoS bandwidth limit methods.
"""

from openstack import exceptions
from openstack.tests.functional import base


class TestQosBandwidthLimitRule(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()
        if not self.operator_cloud:
            self.skipTest("Operator cloud is required for this test")
        if not self.operator_cloud.has_service('network'):
            self.skipTest('Network service not supported by cloud')
        if not self.operator_cloud._has_neutron_extension('qos'):
            self.skipTest('QoS network extension not supported by cloud')

        policy_name = self.getUniqueString('qos_policy')
        self.policy = self.operator_cloud.create_qos_policy(name=policy_name)

        self.addCleanup(self._cleanup_qos_policy)

    def _cleanup_qos_policy(self):
        try:
            self.operator_cloud.delete_qos_policy(self.policy['id'])
        except Exception as e:
            raise exceptions.SDKException(str(e))

    def test_qos_bandwidth_limit_rule_lifecycle(self):
        max_kbps = 1500
        max_burst_kbps = 500
        updated_max_kbps = 2000

        # Create bw limit rule
        rule = self.operator_cloud.create_qos_bandwidth_limit_rule(
            self.policy['id'], max_kbps=max_kbps, max_burst_kbps=max_burst_kbps
        )
        self.assertIn('id', rule)
        self.assertEqual(max_kbps, rule['max_kbps'])
        self.assertEqual(max_burst_kbps, rule['max_burst_kbps'])

        # Now try to update rule
        updated_rule = self.operator_cloud.update_qos_bandwidth_limit_rule(
            self.policy['id'], rule['id'], max_kbps=updated_max_kbps
        )
        self.assertIn('id', updated_rule)
        self.assertEqual(updated_max_kbps, updated_rule['max_kbps'])
        self.assertEqual(max_burst_kbps, updated_rule['max_burst_kbps'])

        # List rules from policy
        policy_rules = self.operator_cloud.list_qos_bandwidth_limit_rules(
            self.policy['id']
        )
        self.assertEqual([updated_rule], policy_rules)

        # Delete rule
        self.operator_cloud.delete_qos_bandwidth_limit_rule(
            self.policy['id'], updated_rule['id']
        )

        # Check if there is no rules in policy
        policy_rules = self.operator_cloud.list_qos_bandwidth_limit_rules(
            self.policy['id']
        )
        self.assertEqual([], policy_rules)

    def test_create_qos_bandwidth_limit_rule_direction(self):
        if not self.operator_cloud._has_neutron_extension(
            'qos-bw-limit-direction'
        ):
            self.skipTest(
                "'qos-bw-limit-direction' network extension "
                "not supported by cloud"
            )
        max_kbps = 1500
        direction = "ingress"
        updated_direction = "egress"

        # Create bw limit rule
        rule = self.operator_cloud.create_qos_bandwidth_limit_rule(
            self.policy['id'], max_kbps=max_kbps, direction=direction
        )
        self.assertIn('id', rule)
        self.assertEqual(max_kbps, rule['max_kbps'])
        self.assertEqual(direction, rule['direction'])

        # Now try to update direction in rule
        updated_rule = self.operator_cloud.update_qos_bandwidth_limit_rule(
            self.policy['id'], rule['id'], direction=updated_direction
        )
        self.assertIn('id', updated_rule)
        self.assertEqual(max_kbps, updated_rule['max_kbps'])
        self.assertEqual(updated_direction, updated_rule['direction'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.functional.shared_file_system.v2 import base


class ShareNetworkSubnetMetadataTest(base.BaseSharedFileSystemTest):
    def setUp(self):
        super().setUp()

        zones = self.operator_cloud.shared_file_system.availability_zones()
        first_zone = next(zones)
        self.user_client = self.user_cloud.shared_file_system

        share_network_name = self.getUniqueString()
        snt = self.user_client.create_share_network(name=share_network_name)
        self.assertIsNotNone(snt)
        self.SHARE_NETWORK_ID = snt.id

        snsb = self.user_client.create_share_network_subnet(
            self.SHARE_NETWORK_ID, availability_zone=first_zone.name
        )
        self.assertIsNotNone(snsb)
        self.SHARE_NETWORK_SUBNET_ID = snsb.id

    def tearDown(self):
        subnet = self.user_client.get_share_network_subnet(
            self.SHARE_NETWORK_ID, self.SHARE_NETWORK_SUBNET_ID
        )
        self.user_client.delete_share_network_subnet(
            self.SHARE_NETWORK_ID,
            self.SHARE_NETWORK_SUBNET_ID,
            ignore_missing=True,
        )
        self.user_client.wait_for_delete(subnet)

        self.user_client.delete_share_network(
            self.SHARE_NETWORK_ID, ignore_missing=True
        )

        super().tearDown()

    def test_share_network_subnet_metadata(self):
        # create
        sot = self.user_client.set_share_network_subnet_metadata(
            self.SHARE_NETWORK_ID,
            self.SHARE_NETWORK_SUBNET_ID,
            foo="bar",
        )
        self.assertEqual(sot['metadata'], {"foo": "bar"})

        # get all metadata
        sot = self.user_client.fetch_share_network_subnet_metadata(
            self.SHARE_NETWORK_ID, self.SHARE_NETWORK_SUBNET_ID
        )
        self.assertEqual('bar', sot["metadata"]["foo"])

        # get metadata item
        sot = self.user_client.fetch_share_network_subnet_metadata_item(
            self.SHARE_NETWORK_ID, self.SHARE_NETWORK_SUBNET_ID, "foo"
        )
        self.assertEqual('bar', sot["metadata"]["foo"])

        # update (merge)
        self.user_client.set_share_network_subnet_metadata(
            self.SHARE_NETWORK_ID,
            self.SHARE_NETWORK_SUBNET_ID,
            new_foo="new_bar",
        )
        sot = self.user_client.fetch_share_network_subnet_metadata(
            self.SHARE_NETWORK_ID, self.SHARE_NETWORK_SUBNET_ID
        )
        self.assertEqual(sot['metadata'], {"foo": "bar", "new_foo": "new_bar"})

        # delete
        self.user_client.delete_share_network_subnet_metadata(
            self.SHARE_NETWORK_ID,
            self.SHARE_NETWORK_SUBNET_ID,
            ["foo"],
        )
        sot = self.user_client.fetch_share_network_subnet_metadata(
            self.SHARE_NETWORK_ID, self.SHARE_NETWORK_SUBNET_ID
        )
        self.assertEqual(sot['metadata'], {"new_foo": "new_bar"})

        # replace with empty
        sot = self.user_client.set_share_network_subnet_metadata(
            self.SHARE_NETWORK_ID,
            self.SHARE_NETWORK_SUBNET_ID,
            replace=True,
        )
        self.assertEqual(sot['metadata'], {})

        sot = self.user_client.fetch_share_network_subnet_metadata(
            self.SHARE_NETWORK_ID, self.SHARE_NETWORK_SUBNET_ID
        )
        self.assertEqual(sot['metadata'], {})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack.block_storage.v3 import service
from openstack.tests.unit import base

EXAMPLE = {
    "binary": "cinder-scheduler",
    "disabled_reason": None,
    "host": "devstack",
    "state": "up",
    "status": "enabled",
    "updated_at": "2017-06-29T05:50:35.000000",
    "zone": "nova",
}


class TestService(base.TestCase):
    def setUp(self):
        super().setUp()
        self.resp = mock.Mock()
        self.resp.body = None  # nothing uses this
        self.resp.json = mock.Mock(return_value={'service': {}})
        self.resp.status_code = 200
        self.resp.headers = {}
        self.sess = mock.Mock()
        self.sess.put = mock.Mock(return_value=self.resp)
        self.sess.default_microversion = '3.0'

    def test_basic(self):
        sot = service.Service()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('services', sot.resources_key)
        self.assertEqual('/os-services', sot.base_path)
        self.assertFalse(sot.allow_commit)
        self.assertTrue(sot.allow_list)
        self.assertFalse(sot.allow_fetch)
        self.assertFalse(sot.allow_delete)

        self.assertDictEqual(
            {
                'binary': 'binary',
                'host': 'host',
                'limit': 'limit',
                'marker': 'marker',
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = service.Service(**EXAMPLE)
        self.assertEqual(EXAMPLE['binary'], sot.binary)
        self.assertEqual(EXAMPLE['binary'], sot.name)
        self.assertEqual(EXAMPLE['disabled_reason'], sot.disabled_reason)
        self.assertEqual(EXAMPLE['host'], sot.host)
        self.assertEqual(EXAMPLE['state'], sot.state)
        self.assertEqual(EXAMPLE['status'], sot.status)
        self.assertEqual(EXAMPLE['zone'], sot.availability_zone)

    def test_enable(self):
        sot = service.Service(**EXAMPLE)

        res = sot.enable(self.sess)
        self.assertIsNotNone(res)

        url = 'os-services/enable'
        body = {
            'binary': 'cinder-scheduler',
            'host': 'devstack',
        }
        self.sess.put.assert_called_with(
            url,
            json=body,
            microversion=self.sess.default_microversion,
        )

    def test_disable(self):
        sot = service.Service(**EXAMPLE)

        res = sot.disable(self.sess)
        self.assertIsNotNone(res)

        url = 'os-services/disable'
        body = {
            'binary': 'cinder-scheduler',
            'host': 'devstack',
        }
        self.sess.put.assert_called_with(
            url,
            json=body,
            microversion=self.sess.default_microversion,
        )

    def test_disable__with_reason(self):
        sot = service.Service(**EXAMPLE)
        reason = 'fencing'

        res = sot.disable(self.sess, reason=reason)

        self.assertIsNotNone(res)

        url = 'os-services/disable-log-reason'
        body = {
            'binary': 'cinder-scheduler',
            'host': 'devstack',
            'disabled_reason': reason,
        }
        self.sess.put.assert_called_with(
            url,
            json=body,
            microversion=self.sess.default_microversion,
        )

    def test_thaw(self):
        sot = service.Service(**EXAMPLE)

        res = sot.thaw(self.sess)
        self.assertIsNotNone(res)

        url = 'os-services/thaw'
        body = {'host': 'devstack'}
        self.sess.put.assert_called_with(
            url,
            json=body,
            microversion=self.sess.default_microversion,
        )

    def test_freeze(self):
        sot = service.Service(**EXAMPLE)

        res = sot.freeze(self.sess)
        self.assertIsNotNone(res)

        url = 'os-services/freeze'
        body = {'host': 'devstack'}
        self.sess.put.assert_called_with(
            url,
            json=body,
            microversion=self.sess.default_microversion,
        )

    def test_set_log_levels(self):
        self.sess.default_microversion = '3.32'
        res = service.Service.set_log_levels(
            self.sess,
            level=service.Level.DEBUG,
            binary=service.Binary.ANY,
            server='foo',
            prefix='cinder.',
        )
        self.assertIsNone(res)

        url = 'os-services/set-log'
        body = {
            'level': service.Level.DEBUG,
            'binary': service.Binary.ANY,
            'server': 'foo',
            'prefix': 'cinder.',
        }
        self.sess.put.assert_called_with(
            url,
            json=body,
            microversion=self.sess.default_microversion,
        )

    def test_get_log_levels(self):
        self.sess.default_microversion = '3.32'
        self.resp.json = mock.Mock(
            return_value={
                'log_levels': [
                    {
                        "binary": "cinder-api",
                        "host": "devstack",
                        "levels": {"cinder.volume.api": "DEBUG"},
                    },
                ],
            },
        )
        res = list(
            service.Service.get_log_levels(
                self.sess,
                binary=service.Binary.ANY,
                server='foo',
                prefix='cinder.',
            )
        )
        self.assertIsNotNone(res)

        url = 'os-services/get-log'
        body = {
            'binary': service.Binary.ANY,
            'server': 'foo',
            'prefix': 'cinder.',
        }
        self.sess.put.assert_called_with(
            url,
            json=body,
            microversion=self.sess.default_microversion,
        )

    @mock.patch(
        'openstack.utils.supports_microversion',
        autospec=True,
        return_value=False,
    )
    def test_failover(self, mock_supports):
        sot = service.Service(**EXAMPLE)

        res = sot.failover(self.sess)
        self.assertIsNotNone(res)

        url = 'os-services/failover_host'
        body = {'host': 'devstack'}
        self.sess.put.assert_called_with(
            url,
            json=body,
            microversion=self.sess.default_microversion,
        )

    @mock.patch(
        'openstack.utils.supports_microversion',
        autospec=True,
        return_value=True,
    )
    def test_failover__with_cluster(self, mock_supports):
        self.sess.default_microversion = '3.26'

        sot = service.Service(**EXAMPLE)

        res = sot.failover(self.sess, cluster='foo', backend_id='bar')
        self.assertIsNotNone(res)

        url = 'os-services/failover'
        body = {
            'host': 'devstack',
            'cluster': 'foo',
            'backend_id': 'bar',
        }
        self.sess.put.assert_called_with(
            url,
            json=body,
            microversion='3.26',
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class Segment(resource.Resource):
    resource_key = 'segment'
    resources_key = 'segments'
    base_path = '/segments'

    _allow_unknown_attrs_in_body = True

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'description',
        'name',
        'network_id',
        'network_type',
        'physical_network',
        'segmentation_id',
        'sort_key',
        'sort_dir',
    )

    # Properties
    #: The segment description.
    description = resource.Body('description')
    #: The segment name.
    name = resource.Body('name')
    #: The ID of the network associated with this segment.
    network_id = resource.Body('network_id')
    #: The type of network associated with this segment, such as
    #: ``flat``, ``geneve``, ``gre``, ``local``, ``vlan`` or ``vxlan``.
    network_type = resource.Body('network_type')
    #: The name of the physical network associated with this segment.
    physical_network = resource.Body('physical_network')
    #: The segmentation ID for this segment. The network type
    #: defines the segmentation model, VLAN ID for ``vlan`` network type
    #: and tunnel ID for ``geneve``, ``gre`` and ``vxlan`` network types.
    #: *Type: int*
    segmentation_id = resource.Body('segmentation_id', type=int)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import uuid

from openstack.dns.v2 import blacklist as _blacklist
from openstack.tests.functional import base


class TestBlackList(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()
        self.require_service('dns')

        # Note: use a unique UUID pattern to avoid test collisions
        self.pattern = rf".*\.test-{uuid.uuid4().hex}.com"
        self.description = self.getUniqueString('blacklist')

    def _delete_blacklist(self, blacklist):
        ret = self.operator_cloud.dns.delete_blacklist(blacklist.id)
        self.assertIsNone(ret)

    def test_blacklist(self):
        # create blacklist
        blacklist = self.operator_cloud.dns.create_blacklist(
            pattern=self.pattern,
            description=self.description,
        )
        self.assertIsNotNone(blacklist.id)
        self.assertIsInstance(blacklist, _blacklist.Blacklist)
        self.assertEqual(self.pattern, blacklist.pattern)
        self.assertEqual(self.description, blacklist.description)
        self.addCleanup(self._delete_blacklist, blacklist)

        # update blacklist
        blacklist = self.operator_cloud.dns.update_blacklist(
            blacklist, pattern=self.pattern, description=self.description
        )
        self.assertIsInstance(blacklist, _blacklist.Blacklist)
        self.assertEqual(self.pattern, blacklist.pattern)
        self.assertEqual(self.description, blacklist.description)

        # get blacklist
        blacklist = self.operator_cloud.dns.get_blacklist(blacklist.id)
        self.assertIsInstance(blacklist, _blacklist.Blacklist)
        self.assertEqual(self.pattern, blacklist.pattern)
        self.assertEqual(self.description, blacklist.description)

        # list all blacklists
        blacklists = list(self.operator_cloud.dns.blacklists())
        self.assertIsInstance(blacklists[0], _blacklist.Blacklist)
        self.assertIn(self.pattern, {x.pattern for x in blacklists})
        self.operator_cloud.dns.delete_blacklist(blacklist.id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import _log
from openstack.compute.v2 import server as server_
from openstack.compute.v2 import server_share as server_share_
from openstack.shared_file_system.v2 import share as share_
from openstack.tests.functional import base as ft_base
from openstack import utils


LOG = _log.setup_logging(__name__)


class TestServerShare(ft_base.BaseFunctionalTest):
    """This test allows testing the functionality through the Shares API
    which enables associating an instance with a share from Manila.
    This test uses a particular Zuul job to meet the prerequisites for
    using this feature. The job will install Devstack:
    1- With the Jammy distribution which contains the appropriate versions
       of Qemu and libvirt.
    2- Install the Manila and Nova services.
    3- Enable the file-backed memory.
    4- Create a test instance and then stop it.

    The association between this test and the job is done through the
    compute_shared_file_system directory which contains this test.
    """

    _wait_for_timeout_key = 'OPENSTACKSDK_FUNC_TEST_TIMEOUT_COMPUTE'

    def setUp(self):
        super().setUp()
        self._set_user_cloud(compute_api_version='2')

        if not self.user_cloud.has_service('compute', '2'):
            self.skipTest('compute service not supported by cloud')

        if not self.user_cloud.has_service('shared-file-system'):
            self.skipTest('shared-file-system service not supported by cloud')

        self.server_name = self.getUniqueString()
        self.share_name = self.getUniqueString()

    def _create_server_and_share(self):
        LOG.debug("Create test server")
        server = self.user_cloud.compute.create_server(
            name=self.server_name,
            flavor_id=self.flavor.id,
            image_id=self.image.id,
            networks='none',
        )
        self.addCleanup(self._delete_server, server)
        self.user_cloud.compute.wait_for_server(
            server,
            wait=self._wait_for_timeout,
        )
        self.assertIsInstance(server, server_.Server)
        self.assertEqual(self.server_name, server.name)

        protocol_supported = (
            "CEPHFS"
            if any(
                "CEPHFS" in p.capabilities["storage_protocol"]
                for p in self.operator_cloud.share.storage_pools()
            )
            else "NFS"
        )

        LOG.debug("Create test share with protocol %s", protocol_supported)
        share = self.user_cloud.share.create_share(
            name=self.share_name,
            size=1,
            share_type="dhss_false",
            share_protocol=protocol_supported,
            description=None,
        )
        self.addCleanup(self._delete_share, share)
        self.user_cloud.share.wait_for_status(
            share,
            status='available',
            wait=self._wait_for_timeout,
        )
        self.assertIsInstance(share, share_.Share)
        self.assertEqual(self.share_name, share.name)

        self.server = server
        self.share = share

    def _delete_server(self, server):
        try:
            LOG.debug("Delete test server")
            # Detach any shares before deleting the server
            try:
                for share in self.user_cloud.compute.share_attachments(server):
                    LOG.debug(
                        "Detaching share %s before server delete",
                        share.share_id,
                    )
                    self.user_cloud.compute.delete_share_attachment(
                        server, share.share_id
                    )
            except Exception:
                LOG.exception(
                    "Failed to detach shares from server %s",
                    server.id,
                )
            self.user_cloud.compute.delete_server(server.id)
            self.user_cloud.compute.wait_for_delete(
                server,
                wait=self._wait_for_timeout,
            )
        except Exception:
            LOG.exception("Failed to delete test server %s", server.id)

    def _delete_share(self, share):
        try:
            LOG.debug("Delete test share")
            self.user_cloud.share.delete_share(share.id)
            self.user_cloud.share.wait_for_delete(
                share,
                wait=self._wait_for_timeout,
            )
        except Exception:
            LOG.exception("Failed to delete test share %s", share.id)

    def _wait_for_share_status(self, server, expected_status, count=None):
        """Wait for share attachments to reach the expected status."""
        for _ in utils.iterate_timeout(
            self._wait_for_timeout,
            message=f"Wait for share attachment to be {expected_status}",
        ):
            server_shares = list(
                self.user_cloud.compute.share_attachments(server)
            )
            self.assertEqual(1, len(server_shares))
            self.assertIsInstance(
                server_shares[0],
                server_share_.ShareMapping,
            )

            if server_shares[0].status == expected_status:
                return server_shares[0]

    def _wait_for_no_shares(self, server):
        """Wait until all share attachments are removed."""
        for _ in utils.iterate_timeout(
            self._wait_for_timeout,
            message="Wait for deletion of share attachment",
        ):
            server_shares = list(
                self.user_cloud.compute.share_attachments(server)
            )
            if len(server_shares) == 0:
                return

    def test_server_share(self):
        self._create_server_and_share()

        # Stop server
        LOG.debug("Stop test server to attach share")
        self.user_cloud.compute.stop_server(self.server.id)
        self.user_cloud.compute.wait_for_server(
            self.server,
            status='SHUTOFF',
            wait=self._wait_for_timeout,
        )

        # Create the server attachment
        LOG.debug("Attach share to test server")
        server_share = self.user_cloud.compute.create_share_attachment(
            self.server.id,
            self.share.id,
        )
        self.assertIsInstance(
            server_share,
            server_share_.ShareMapping,
        )

        # List all attached server shares (there should only be one)
        share_mapping = self._wait_for_share_status(
            self.server,
            'inactive',
        )
        self.assertEqual(share_mapping.status, 'inactive')

        # Retrieve details of the server share
        server_share = self.user_cloud.compute.get_share_attachment(
            self.server,
            self.share.id,
        )
        self.assertIsInstance(
            server_share,
            server_share_.ShareMapping,
        )
        self.assertEqual(server_share.status, 'inactive')

        # Detach the server share
        LOG.debug("Detach server share")
        result = self.user_cloud.compute.delete_share_attachment(
            self.server,
            self.share.id,
        )
        self.assertIsNone(result)
        self._wait_for_no_shares(self.server)

        # Create the server share with a tag
        LOG.debug("Attach server share with a tag")
        server_share = self.user_cloud.compute.create_share_attachment(
            self.server.id, self.share.id, tag='mytag'
        )
        self.assertIsInstance(
            server_share,
            server_share_.ShareMapping,
        )

        # Retrieve details of the server share and check tag
        share_mapping = self._wait_for_share_status(
            self.server,
            'inactive',
        )
        self.assertEqual(share_mapping.status, 'inactive')
        self.assertEqual(share_mapping.tag, 'mytag')

        # Start server
        LOG.debug("Start test server")
        self.user_cloud.compute.start_server(self.server.id)
        try:
            self.user_cloud.compute.wait_for_server(
                self.server,
                status='ACTIVE',
                wait=self._wait_for_timeout,
            )
        except Exception:
            # Fetch server details to capture fault information
            srv = self.user_cloud.compute.get_server(self.server.id)
            LOG.error(
                "Server failed to reach ACTIVE. Status: %s, Fault: %s",
                srv.status,
                srv.fault,
            )
            raise

        # Retrieve details of the server share after power on
        share_mapping = self._wait_for_share_status(
            self.server,
            'active',
        )
        self.assertEqual(share_mapping.status, 'active')
        self.assertEqual(share_mapping.tag, 'mytag')

        # Stop server
        LOG.debug("Stop test server")
        self.user_cloud.compute.stop_server(self.server.id)
        self.user_cloud.compute.wait_for_server(
            self.server,
            status='SHUTOFF',
            wait=self._wait_for_timeout,
        )

        # Detach the server share
        LOG.debug("Detach server share")
        result = self.user_cloud.compute.delete_share_attachment(
            self.server,
            self.share.id,
        )
        self.assertIsNone(result)
        self._wait_for_no_shares(self.server)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import Callable, Iterable
from typing import Any, cast

from keystoneauth1 import adapter
import requests

from openstack import exceptions
from openstack import resource
from openstack import utils


class Type(resource.Resource):
    resource_key = "volume_type"
    resources_key = "volume_types"
    base_path = "/types"

    # capabilities
    allow_fetch = True
    allow_create = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters("is_public")

    # Properties
    #: A dict of extra specifications. "capabilities" is a usual key.
    extra_specs = resource.Body("extra_specs", type=dict)
    #: a private volume-type. *Type: bool*
    is_public = resource.Body('os-volume-type-access:is_public', type=bool)

    def _extra_specs(
        self,
        method: Callable[..., requests.Response],
        key: str | None = None,
        delete: bool = False,
        extra_specs: dict[str, str] | None = None,
    ) -> dict[str, Any] | None:
        extra_specs = extra_specs or {}
        for k, v in extra_specs.items():
            if not isinstance(v, str):
                raise ValueError(
                    f"The value for {k} ({v}) must be a text string"
                )

        if key is not None:
            url = utils.urljoin(self.base_path, self.id, "extra_specs", key)
        else:
            url = utils.urljoin(self.base_path, self.id, "extra_specs")

        kwargs = {}
        if extra_specs:
            kwargs["json"] = {"extra_specs": extra_specs}

        response = method(url, headers={}, **kwargs)

        exceptions.raise_from_response(response)
        return response.json() if not delete else None

    def set_extra_specs(
        self, session: adapter.Adapter, **extra_specs: str
    ) -> dict[str, Any]:
        """Update extra specs.

        This call will replace only the extra_specs with the same keys
        given here.  Other keys will not be modified.

        :param session: The session to use for making this request.
        :param extra_specs: Key/value extra_specs pairs to be update on
            this volume type. All keys and values.
        :returns: The updated extra specs.
        """
        if not extra_specs:
            return dict()

        result = self._extra_specs(session.post, extra_specs=extra_specs)
        assert result is not None
        return cast(dict[str, Any], result["extra_specs"])

    def delete_extra_specs(
        self, session: adapter.Adapter, keys: Iterable[str]
    ) -> None:
        """Delete extra specs.

        .. note::

            This method will do a HTTP DELETE request for every key in keys.

        :param session: The session to use for this request.
        :param keys: The keys to delete.
        :returns: ``None``
        """
        for key in keys:
            self._extra_specs(session.delete, key=key, delete=True)

    def get_private_access(
        self, session: adapter.Adapter
    ) -> list[dict[str, Any]]:
        """List projects with private access to the volume type.

        :param session: The session to use for making this request.
        :returns: The volume type access response.
        """
        url = utils.urljoin(self.base_path, self.id, "os-volume-type-access")
        resp = session.get(url)

        exceptions.raise_from_response(resp)

        return cast(
            list[dict[str, Any]], resp.json().get("volume_type_access", [])
        )

    def add_private_access(
        self, session: adapter.Adapter, project_id: str
    ) -> None:
        """Add project access from the volume type.

        :param session: The session to use for making this request.
        :param project_id: The project to add access for.
        """
        url = utils.urljoin(self.base_path, self.id, "action")
        body = {"addProjectAccess": {"project": project_id}}

        resp = session.post(url, json=body)

        exceptions.raise_from_response(resp)

    def remove_private_access(
        self, session: adapter.Adapter, project_id: str
    ) -> None:
        """Remove project access from the volume type.

        :param session: The session to use for making this request.
        :param project_id: The project to remove access for.
        """
        url = utils.urljoin(self.base_path, self.id, "action")
        body = {"removeProjectAccess": {"project": project_id}}

        resp = session.post(url, json=body)

        exceptions.raise_from_response(resp)


class TypeEncryption(resource.Resource):
    resource_key = "encryption"
    resources_key = "encryption"
    base_path = "/types/%(volume_type_id)s/encryption"

    # capabilities
    allow_fetch = True
    allow_create = True
    allow_delete = True
    allow_list = False
    allow_commit = True

    # Properties
    #: The encryption algorithm or mode.
    cipher = resource.Body("cipher")
    #: Notional service where encryption is performed.
    control_location = resource.Body("control_location")
    #: The date and time when the resource was created.
    created_at = resource.Body("created_at")
    #: The resource is deleted or not.
    deleted = resource.Body("deleted")
    #: The date and time when the resource was deleted.
    deleted_at = resource.Body("deleted_at")
    #: A ID representing this type.
    encryption_id = resource.Body("encryption_id", alternate_id=True)
    #: The Size of encryption key.
    key_size = resource.Body("key_size")
    #: The class that provides encryption support.
    provider = resource.Body("provider")
    #: The date and time when the resource was updated.
    updated_at = resource.Body("updated_at")
    #: The ID of the Volume Type.
    volume_type_id = resource.URI("volume_type_id")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.identity.v3 import _proxy as _identity_v3
from openstack.tests.functional import base
from openstack import utils


class BaseIdentityTest(base.BaseFunctionalTest):
    admin_identity_client: _identity_v3.Proxy
    system_admin_identity_client: _identity_v3.Proxy

    def setUp(self):
        super().setUp()
        if not self.operator_cloud.has_service('identity', '3'):
            self.skipTest('identity service not supported by cloud')

        self.admin_identity_client = utils.ensure_service_version(
            self.operator_cloud.identity, '3'
        )
        self.system_admin_identity_client = utils.ensure_service_version(
            self.system_admin_cloud.identity, '3'
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneauth1 import adapter

from openstack.identity.v3 import domain
from openstack.identity.v3 import group
from openstack.identity.v3 import role
from openstack.identity.v3 import user
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'description': '1',
    'enabled': True,
    'id': IDENTIFIER,
    'links': {'self': f'http://example.com/identity/v3/domains/{IDENTIFIER}'},
    'name': '4',
    'options': {'foo': 'bar'},
}


class TestDomain(base.TestCase):
    def setUp(self):
        super().setUp()
        self.sess = mock.Mock(spec=adapter.Adapter)
        self.sess.default_microversion = 1
        self.sess._get_connection = mock.Mock(return_value=self.cloud)
        self.good_resp = mock.Mock()
        self.good_resp.body = None
        self.good_resp.json = mock.Mock(return_value=self.good_resp.body)
        self.good_resp.status_code = 204

        self.bad_resp = mock.Mock()
        self.bad_resp.body = None
        self.bad_resp.json = mock.Mock(return_value=self.bad_resp.body)
        self.bad_resp.status_code = 401

    def test_basic(self):
        sot = domain.Domain()
        self.assertEqual('domain', sot.resource_key)
        self.assertEqual('domains', sot.resources_key)
        self.assertEqual('/domains', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertEqual('PATCH', sot.commit_method)

        self.assertDictEqual(
            {
                'name': 'name',
                'is_enabled': 'enabled',
                'limit': 'limit',
                'marker': 'marker',
                'tags': 'tags',
                'any_tags': 'tags-any',
                'not_tags': 'not-tags',
                'not_any_tags': 'not-tags-any',
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = domain.Domain(**EXAMPLE)
        self.assertEqual(EXAMPLE['description'], sot.description)
        self.assertTrue(sot.is_enabled)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['links'], sot.links)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertDictEqual(EXAMPLE['options'], sot.options)

    def test_assign_role_to_user_good(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.good_resp
        self.sess.put = mock.Mock(return_value=resp)

        self.assertTrue(
            sot.assign_role_to_user(
                self.sess, user.User(id='1'), role.Role(id='2'), False
            )
        )

        self.sess.put.assert_called_with('domains/IDENTIFIER/users/1/roles/2')

    def test_assign_inherited_role_to_user_good(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.good_resp
        self.sess.put = mock.Mock(return_value=resp)

        self.assertTrue(
            sot.assign_role_to_user(
                self.sess, user.User(id='1'), role.Role(id='2'), True
            )
        )

        self.sess.put.assert_called_with(
            'OS-INHERIT/domains/IDENTIFIER/users/1/roles/2/inherited_to_projects'
        )

    def test_assign_role_to_user_bad(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.bad_resp
        self.sess.put = mock.Mock(return_value=resp)

        self.assertFalse(
            sot.assign_role_to_user(
                self.sess, user.User(id='1'), role.Role(id='2'), False
            )
        )

    def test_validate_user_has_role_good(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.good_resp
        self.sess.head = mock.Mock(return_value=resp)

        self.assertTrue(
            sot.validate_user_has_role(
                self.sess, user.User(id='1'), role.Role(id='2'), False
            )
        )

        self.sess.head.assert_called_with('domains/IDENTIFIER/users/1/roles/2')

    def test_validate_user_has_inherited_role_good(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.good_resp
        self.sess.head = mock.Mock(return_value=resp)

        self.assertTrue(
            sot.validate_user_has_role(
                self.sess, user.User(id='1'), role.Role(id='2'), True
            )
        )

        self.sess.head.assert_called_with(
            'OS-INHERIT/domains/IDENTIFIER/users/1/roles/2/inherited_to_projects'
        )

    def test_validate_user_has_role_bad(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.bad_resp
        self.sess.head = mock.Mock(return_value=resp)

        self.assertFalse(
            sot.validate_user_has_role(
                self.sess, user.User(id='1'), role.Role(id='2'), False
            )
        )

    def test_unassign_role_from_user_good(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.good_resp
        self.sess.delete = mock.Mock(return_value=resp)

        self.assertTrue(
            sot.unassign_role_from_user(
                self.sess, user.User(id='1'), role.Role(id='2'), False
            )
        )

        self.sess.delete.assert_called_with(
            'domains/IDENTIFIER/users/1/roles/2'
        )

    def test_unassign_inherited_role_from_user_good(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.good_resp
        self.sess.delete = mock.Mock(return_value=resp)

        self.assertTrue(
            sot.unassign_role_from_user(
                self.sess, user.User(id='1'), role.Role(id='2'), True
            )
        )

        self.sess.delete.assert_called_with(
            'OS-INHERIT/domains/IDENTIFIER/users/1/roles/2/inherited_to_projects'
        )

    def test_unassign_role_from_user_bad(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.bad_resp
        self.sess.delete = mock.Mock(return_value=resp)

        self.assertFalse(
            sot.unassign_role_from_user(
                self.sess, user.User(id='1'), role.Role(id='2'), False
            )
        )

    def test_assign_role_to_group_good(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.good_resp
        self.sess.put = mock.Mock(return_value=resp)

        self.assertTrue(
            sot.assign_role_to_group(
                self.sess, group.Group(id='1'), role.Role(id='2'), False
            )
        )

        self.sess.put.assert_called_with('domains/IDENTIFIER/groups/1/roles/2')

    def test_assign_inherited_role_to_group_good(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.good_resp
        self.sess.put = mock.Mock(return_value=resp)

        self.assertTrue(
            sot.assign_role_to_group(
                self.sess, group.Group(id='1'), role.Role(id='2'), True
            )
        )

        self.sess.put.assert_called_with(
            'OS-INHERIT/domains/IDENTIFIER/groups/1/roles/2/inherited_to_projects'
        )

    def test_assign_role_to_group_bad(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.bad_resp
        self.sess.put = mock.Mock(return_value=resp)

        self.assertFalse(
            sot.assign_role_to_group(
                self.sess, group.Group(id='1'), role.Role(id='2'), False
            )
        )

    def test_validate_group_has_role_good(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.good_resp
        self.sess.head = mock.Mock(return_value=resp)

        self.assertTrue(
            sot.validate_group_has_role(
                self.sess, group.Group(id='1'), role.Role(id='2'), False
            )
        )

        self.sess.head.assert_called_with(
            'domains/IDENTIFIER/groups/1/roles/2'
        )

    def test_validate_group_has_inherited_role_good(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.good_resp
        self.sess.head = mock.Mock(return_value=resp)

        self.assertTrue(
            sot.validate_group_has_role(
                self.sess, group.Group(id='1'), role.Role(id='2'), True
            )
        )

        self.sess.head.assert_called_with(
            'OS-INHERIT/domains/IDENTIFIER/groups/1/roles/2/inherited_to_projects'
        )

    def test_validate_group_has_role_bad(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.bad_resp
        self.sess.head = mock.Mock(return_value=resp)

        self.assertFalse(
            sot.validate_group_has_role(
                self.sess, group.Group(id='1'), role.Role(id='2'), False
            )
        )

    def test_unassign_role_from_group_good(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.good_resp
        self.sess.delete = mock.Mock(return_value=resp)

        self.assertTrue(
            sot.unassign_role_from_group(
                self.sess, group.Group(id='1'), role.Role(id='2'), False
            )
        )

        self.sess.delete.assert_called_with(
            'domains/IDENTIFIER/groups/1/roles/2'
        )

    def test_unassign_inherited_role_from_group_good(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.good_resp
        self.sess.delete = mock.Mock(return_value=resp)

        self.assertTrue(
            sot.unassign_role_from_group(
                self.sess, group.Group(id='1'), role.Role(id='2'), True
            )
        )

        self.sess.delete.assert_called_with(
            'OS-INHERIT/domains/IDENTIFIER/groups/1/roles/2/inherited_to_projects'
        )

    def test_unassign_role_from_group_bad(self):
        sot = domain.Domain(**EXAMPLE)
        resp = self.bad_resp
        self.sess.delete = mock.Mock(return_value=resp)

        self.assertFalse(
            sot.unassign_role_from_group(
                self.sess, group.Group(id='1'), role.Role(id='2'), False
            )
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Self
import warnings

from keystoneauth1 import adapter

from openstack import exceptions
from openstack import resource
from openstack import utils
from openstack import warnings as os_warnings


class Hypervisor(resource.Resource):
    resource_key = 'hypervisor'
    resources_key = 'hypervisors'
    base_path = '/os-hypervisors'

    # capabilities
    allow_fetch = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'hypervisor_hostname_pattern', 'with_servers'
    )

    # Lot of attributes are dropped in 2.88
    _max_microversion = '2.88'

    # Properties
    #: Information about the hypervisor's CPU. Up to 2.28 it was string.
    cpu_info = resource.Body('cpu_info')
    #: IP address of the host
    host_ip = resource.Body('host_ip')
    #: The type of hypervisor
    hypervisor_type = resource.Body('hypervisor_type')
    #: Version of the hypervisor
    hypervisor_version = resource.Body('hypervisor_version')
    #: Name of hypervisor
    name = resource.Body('hypervisor_hostname')
    #: Service details
    service_details = resource.Body('service', type=dict)
    #: List of Servers
    servers = resource.Body('servers', type=list, list_type=dict)
    #: State of hypervisor
    state = resource.Body('state')
    #: Status of hypervisor
    status = resource.Body('status')
    #: The total uptime of the hypervisor and information about average load.
    #: This attribute is set only when querying uptime explicitly.
    uptime = resource.Body('uptime')

    # Attributes deprecated with 2.88
    #: Measurement of the hypervisor's current workload
    current_workload = resource.Body('current_workload', deprecated=True)
    #: Disk space available to the scheduler
    disk_available = resource.Body("disk_available_least", deprecated=True)
    #: The amount, in gigabytes, of local storage used
    local_disk_used = resource.Body('local_gb_used', deprecated=True)
    #: The amount, in gigabytes, of the local storage device
    local_disk_size = resource.Body('local_gb', deprecated=True)
    #: The amount, in gigabytes, of free space on the local storage device
    local_disk_free = resource.Body('free_disk_gb', deprecated=True)
    #: The amount, in megabytes, of memory
    memory_used = resource.Body('memory_mb_used', deprecated=True)
    #: The amount, in megabytes, of total memory
    memory_size = resource.Body('memory_mb', deprecated=True)
    #: The amount, in megabytes, of available memory
    memory_free = resource.Body('free_ram_mb', deprecated=True)
    #: Count of the running virtual machines
    running_vms = resource.Body('running_vms', deprecated=True)
    #: Count of the VCPUs in use
    vcpus_used = resource.Body('vcpus_used', deprecated=True)
    #: Count of all VCPUs
    vcpus = resource.Body('vcpus', deprecated=True)

    def get_uptime(self, session: adapter.Adapter) -> Self:
        """Get uptime information for the hypervisor

        Updates uptime attribute of the hypervisor object
        """
        warnings.warn(
            "This call is deprecated and is only available until Nova 2.88",
            os_warnings.LegacyAPIWarning,
        )
        if utils.supports_microversion(session, '2.88'):
            raise exceptions.SDKException(
                'Hypervisor.get_uptime is not supported anymore'
            )
        url = utils.urljoin(self.base_path, self.id, 'uptime')
        microversion = self._get_microversion(session)
        response = session.get(url, microversion=microversion)
        self._translate_response(response)
        return self


HypervisorDetail = Hypervisor



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import os
from typing import Any, ClassVar, Literal, overload
from collections.abc import Callable, Generator
import warnings

import requests

from openstack import exceptions as exc
from openstack.image.v1 import image as _image
from openstack import proxy
from openstack import resource
from openstack import utils
from openstack import warnings as os_warnings


def _get_name_and_filename(
    name: str, image_format: str
) -> tuple[str, str | None]:
    # See if name points to an existing file
    if os.path.exists(name):
        # Neat. Easy enough
        return os.path.splitext(os.path.basename(name))[0], name

    # Try appending the disk format
    name_with_ext = '.'.join((name, image_format))
    if os.path.exists(name_with_ext):
        return os.path.basename(name), name_with_ext

    return name, None


class Proxy(proxy.Proxy):
    api_version: ClassVar[Literal['1']] = '1'

    retriable_status_codes = [503]

    _IMAGE_MD5_KEY = 'owner_specified.openstack.md5'
    _IMAGE_SHA256_KEY = 'owner_specified.openstack.sha256'
    _IMAGE_OBJECT_KEY = 'owner_specified.openstack.object'

    # NOTE(shade) shade keys were owner_specified.shade.md5 - we need to add
    #             those to freshness checks so that a shade->sdk transition
    #             doesn't result in a re-upload
    _SHADE_IMAGE_MD5_KEY = 'owner_specified.shade.md5'
    _SHADE_IMAGE_SHA256_KEY = 'owner_specified.shade.sha256'
    _SHADE_IMAGE_OBJECT_KEY = 'owner_specified.shade.object'

    # ====== IMAGES ======
    def create_image(
        self,
        name: str,
        filename: str | None = None,
        container: str | None = None,
        md5: str | None = None,
        sha256: str | None = None,
        disk_format: str | None = None,
        container_format: str | None = None,
        disable_vendor_agent: bool = True,
        allow_duplicates: bool = False,
        meta: dict[str, Any] | None = None,
        data: bytes | None = None,
        validate_checksum: bool = False,
        tags: list[str] | None = None,
        **kwargs: Any,
    ) -> _image.Image:
        """Create an image and optionally upload data.

        Create a new image. If ``filename`` or ``data`` are provided, it will
        also upload data to this image.

        :param name: Name of the image to create. If it is a path name
            of an image, the name will be constructed from the extensionless
            basename of the path.
        :param filename: The path to the file to upload, if needed.
            (optional, defaults to None)
        :param data: Image data (string or file-like object). It is mutually
            exclusive with filename
        :param container: Name of the container in swift where images
            should be uploaded for import if the cloud requires such a thing.
            (optional, defaults to 'images')
        :param md5: md5 sum of the image file. If not given, an md5 will
            be calculated.
        :param sha256: sha256 sum of the image file. If not given, an md5
            will be calculated.
        :param disk_format: The disk format the image is in. (optional,
            defaults to the os-client-config config value for this cloud)
        :param container_format: The container format the image is in.
            (optional, defaults to the os-client-config config value for this
            cloud)
        :param tags: List of tags for this image. Each tag is a string
            of at most 255 chars.
        :param disable_vendor_agent: Whether or not to append metadata
            flags to the image to inform the cloud in question to not expect a
            vendor agent to be runing. (optional, defaults to True)
        :param allow_duplicates: If true, skips checks that enforce unique
            image name. (optional, defaults to False)
        :param meta: A dict of key/value pairs to use for metadata that
            bypasses automatic type conversion.
        :param validate_checksum: If true and cloud returns checksum,
            compares return value with the one calculated or passed into this
            call. If value does not match - raises exception. Default is
            'false'

        Additional kwargs will be passed to the image creation as additional
        metadata for the image and will have all values converted to string
        except for min_disk, min_ram, size and virtual_size which will be
        converted to int.

        If you are sure you have all of your data types correct or have an
        advanced need to be explicit, use meta. If you are just a normal
        consumer, using kwargs is likely the right choice.

        If a value is in meta and kwargs, meta wins.

        :returns: The results of image creation
        :raises: SDKException if there are problems uploading
        """
        # these were previously provided for API (method) compatibility; that
        # was a bad idea
        if (
            'use_import' in kwargs
            or 'stores' in kwargs
            or 'all_stores' in kwargs
            or 'all_stores_must_succeed' in kwargs
        ):
            raise exc.InvalidRequest(
                "Glance v1 does not support stores or image import"
            )

        # silently ignore these; they were never supported and were only given
        # for API (method) compatibility
        kwargs.pop('wait')
        kwargs.pop('timeout')

        if container is None:
            container = self._connection._OBJECT_AUTOCREATE_CONTAINER

        if not meta:
            meta = {}

        if not disk_format:
            disk_format = self._connection.config.config['image_format']

        if not container_format:
            # https://docs.openstack.org/image-guide/image-formats.html
            container_format = 'bare'

        if data and filename:
            raise exc.SDKException(
                'Passing filename and data simultaneously is not supported'
            )

        # If there is no filename, see if name is actually the filename
        if not filename and not data:
            name, filename = _get_name_and_filename(
                name,
                self._connection.config.config['image_format'],
            )

        if validate_checksum and data and not isinstance(data, bytes):
            raise exc.SDKException(
                'Validating checksum is not possible when data is not a '
                'direct binary object'
            )

        if not (md5 or sha256) and validate_checksum:
            if filename:
                md5, sha256 = utils._get_file_hashes(filename)
            elif data and isinstance(data, bytes):
                md5, sha256 = utils._calculate_data_hashes(data)

        if allow_duplicates:
            current_image = None
        else:
            current_image = self.find_image(name)
            if current_image:
                # NOTE(pas-ha) 'properties' may be absent or be None
                props = current_image.get('properties') or {}
                md5_key = props.get(
                    self._IMAGE_MD5_KEY,
                    props.get(self._SHADE_IMAGE_MD5_KEY, ''),
                )
                sha256_key = props.get(
                    self._IMAGE_SHA256_KEY,
                    props.get(self._SHADE_IMAGE_SHA256_KEY, ''),
                )
                up_to_date = utils._hashes_up_to_date(
                    md5=md5,
                    sha256=sha256,
                    md5_key=md5_key,
                    sha256_key=sha256_key,
                )
                if up_to_date:
                    self.log.debug(
                        "image %(name)s exists and is up to date",
                        {'name': name},
                    )
                    return current_image
                else:
                    self.log.debug(
                        "image %(name)s exists, but contains different "
                        "checksums. Updating.",
                        {'name': name},
                    )

        if disable_vendor_agent:
            kwargs.update(
                self._connection.config.config['disable_vendor_agent']
            )

        # If a user used the v1 calling format, they will have
        # passed a dict called properties along
        properties = kwargs.pop('properties', {})
        properties[self._IMAGE_MD5_KEY] = md5 or ''
        properties[self._IMAGE_SHA256_KEY] = sha256 or ''
        properties[self._IMAGE_OBJECT_KEY] = '/'.join([container, name])
        kwargs.update(properties)
        image_kwargs: dict[str, Any] = {'properties': kwargs}
        if disk_format:
            image_kwargs['disk_format'] = disk_format
        if container_format:
            image_kwargs['container_format'] = container_format
        if tags:
            image_kwargs['tags'] = tags

        if filename or data:
            image = self._upload_image(
                name,
                filename=filename,
                data=data,
                meta=meta,
                validate_checksum=validate_checksum,
                **image_kwargs,
            )
        else:
            image = self._create(_image.Image, name=name, **kwargs)

        return image  # type: ignore[no-any-return]

    def upload_image(self, **attrs: Any) -> _image.Image:
        """Upload a new image from attributes

        .. warning:
          This method is deprecated - and also doesn't work very well.
          Please stop using it immediately and switch to
          `create_image`.

        :param attrs: Keyword arguments which will be used to create
            a :class:`~openstack.image.v1.image.Image`,
            comprised of the properties on the Image class.

        :returns: The results of image creation
        """
        warnings.warn(
            "upload_image is deprecated. Use create_image instead.",
            os_warnings.RemovedInSDK50Warning,
        )
        return self._create(_image.Image, **attrs)

    def _upload_image(
        self,
        name: str,
        filename: str | None,
        data: Any,
        meta: dict[str, Any] | None,
        **image_kwargs: Any,
    ) -> Any:
        if filename and not data:
            image_data = open(filename, 'rb')
        else:
            image_data = data
        image_kwargs['properties'].update(meta)
        image_kwargs['name'] = name

        # TODO(mordred) Convert this to use image Resource
        image = self._connection._get_and_munchify(  # type: ignore[no-untyped-call]
            'image', self.post('/images', json=image_kwargs)
        )
        checksum = image_kwargs['properties'].get(self._IMAGE_MD5_KEY, '')

        try:
            # Let us all take a brief moment to be grateful that this
            # is not actually how OpenStack APIs work anymore
            headers = {
                'x-glance-registry-purge-props': 'false',
            }
            if checksum:
                headers['x-image-meta-checksum'] = checksum

            image = self._connection._get_and_munchify(  # type: ignore[no-untyped-call]
                'image',
                self.put(
                    f'/images/{image.id}',
                    headers=headers,
                    data=image_data,
                ),
            )
        except exc.HttpException:
            self.log.debug("Deleting failed upload of image %s", name)
            try:
                self.delete(f'/images/{image.id}')
            except exc.HttpException:
                # We're just trying to clean up - if it doesn't work - shrug
                self.log.warning(
                    "Failed deleting image after we failed uploading it.",
                    exc_info=True,
                )
            raise
        return image

    def _existing_image(self, **kwargs: Any) -> _image.Image:
        return _image.Image.existing(connection=self._connection, **kwargs)

    def delete_image(
        self, image: str | _image.Image, ignore_missing: bool = True
    ) -> None:
        """Delete an image

        :param image: The value can be either the ID of an image or a
            :class:`~openstack.image.v1.image.Image` instance.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the image does not exist.
            When set to ``True``, no exception will be set when
            attempting to delete a nonexistent image.

        :returns: ``None``
        """
        self._delete(_image.Image, image, ignore_missing=ignore_missing)

    @overload
    def find_image(
        self,
        name_or_id: str,
        ignore_missing: Literal[False],
    ) -> _image.Image: ...

    @overload
    def find_image(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
    ) -> _image.Image | None: ...

    def find_image(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
    ) -> _image.Image | None:
        """Find a single image

        :param name_or_id: The name or ID of a image.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the resource does not exist.
            When set to ``True``, None will be returned when
            attempting to find a nonexistent resource.
        :returns: One :class:`~openstack.image.v1.image.Image` or None
        """
        return self._find(
            _image.Image, name_or_id, ignore_missing=ignore_missing
        )

    def get_image(self, image: str | _image.Image) -> _image.Image:
        """Get a single image

        :param image: The value can be the ID of an image or a
            :class:`~openstack.image.v1.image.Image` instance.

        :returns: One :class:`~openstack.image.v1.image.Image`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(_image.Image, image)

    def images(self, **query: Any) -> Generator[_image.Image, None, None]:
        """Return a generator of images

        :param query: Optional query parameters to be sent to limit
            the resources being returned.

        :returns: A generator of image objects
        """
        return self._list(_image.Image, base_path='/images/detail', **query)

    def update_image(
        self, image: str | _image.Image, **attrs: Any
    ) -> _image.Image:
        """Update a image

        :param image: Either the ID of a image or a
            :class:`~openstack.image.v1.image.Image` instance.
        :param attrs: The attributes to update on the image represented
            by ``image``.

        :returns: The updated image
        """
        return self._update(_image.Image, image, **attrs)

    def download_image(
        self,
        image: str | _image.Image,
        stream: bool = False,
        output: str | io.IOBase | None = None,
        chunk_size: int = 1024 * 1024,
    ) -> requests.Response:
        """Download an image

        This will download an image to memory when ``stream=False``, or allow
        streaming downloads using an iterator when ``stream=True``.
        For examples of working with streamed responses, see
        :ref:`download_image-stream-true`.

        :param image: The value can be either the ID of an image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param stream: When ``True``, return a :class:`requests.Response`
            instance allowing you to iterate over the
            response data stream instead of storing its entire
            contents in memory. See
            :meth:`requests.Response.iter_content` for more
            details. *NOTE*: If you do not consume
            the entirety of the response you must explicitly
            call :meth:`requests.Response.close` or otherwise
            risk inefficiencies with the ``requests``
            library's handling of connections.

            When ``False``, return the entire contents of the response.
        :param output: Either a file object or a path to store data into.
        :param chunk_size: size in bytes to read from the wire and buffer
            at one time. Defaults to 1024 * 1024 = 1 MiB

        :returns: When output is not given - the bytes comprising the given
            Image when stream is False, otherwise a :class:`requests.Response`
            instance. When output is given - a
            :class:`~openstack.image.v2.image.Image` instance.
        """

        image = self._get_resource(_image.Image, image)

        return image.download(
            self,
            stream=stream,
            output=output,
            chunk_size=chunk_size,
        )

    def _update_image_properties(
        self,
        image: _image.Image,
        meta: dict[str, Any],
        properties: dict[str, Any],
    ) -> bool:
        properties.update(meta)
        img_props = {}
        for k, v in iter(properties.items()):
            if image.properties.get(k, None) != v:
                img_props[f'x-image-meta-{k}'] = v
        if not img_props:
            return False
        self.put(f'/images/{image.id}', headers=img_props)
        return True

    def update_image_properties(
        self,
        image: str | _image.Image | None = None,
        meta: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> bool:
        """
        Update the properties of an existing image.

        :param image: Name or id of an image or an Image object.
        :param meta: A dict of key/value pairs to use for metadata that
            bypasses automatic type conversion.

        Additional kwargs will be passed to the image creation as additional
        metadata for the image and will have all values converted to string
        except for min_disk, min_ram, size and virtual_size which will be
        converted to int.
        """

        if isinstance(image, str):
            image = self._connection.get_image(image)  # type: ignore[no-untyped-call]

        if not meta:
            meta = {}

        img_props = {}
        for k, v in iter(kwargs.items()):
            if v and k in ['ramdisk', 'kernel']:
                v = self._connection.get_image_id(v)  # type: ignore[no-untyped-call]
                k = f'{k}_id'
            img_props[k] = v

        return self._update_image_properties(image, meta, img_props)  # type: ignore[arg-type]

    # ========== Utilities ==========

    def wait_for_status(
        self,
        res: resource.ResourceT,
        status: str,
        failures: list[str] | None = None,
        interval: int | float | None = 2,
        wait: int | None = None,
        attribute: str = 'status',
        callback: Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for the resource to be in a particular status.

        :param session: The session to use for making this request.
        :param resource: The resource to wait on to reach the status. The
            resource must have a status attribute specified via ``attribute``.
        :param status: Desired status of the resource.
        :param failures: Statuses that would indicate the transition
            failed such as 'ERROR'. Defaults to ['ERROR'].
        :param interval: Number of seconds to wait between checks.
        :param wait: Maximum number of seconds to wait for transition.
            Set to ``None`` to wait forever.
        :param attribute: Name of the resource attribute that contains the
            status.
        :param callback: A callback function. This will be called with a single
            value, progress. This is API specific but is generally a percentage
            value from 0-100.

        :returns: The updated resource.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if the
            transition to status failed to occur in ``wait`` seconds.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if the resource
            transitioned to one of the states in ``failures``.
        :raises: :class:`~AttributeError` if the resource does not have a
            ``status`` attribute
        """
        return resource.wait_for_status(
            self, res, status, failures, interval, wait, attribute, callback
        )

    def wait_for_delete(
        self,
        res: resource.ResourceT,
        interval: int | float | None = 2,
        wait: int | None = 120,
        callback: Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for a resource to be deleted.

        :param res: The resource to wait on to be deleted.
        :param interval: Number of seconds to wait before to consecutive
            checks.
        :param wait: Maximum number of seconds to wait before the change.
        :param callback: A callback function. This will be called with a single
            value, progress, which is a percentage value from 0-100.

        :returns: The resource is returned on success.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
            to delete failed to occur in the specified seconds.
        """
        return resource.wait_for_delete(self, res, interval, wait, callback)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack.orchestration.v1 import template
from openstack import resource
from openstack.tests.unit import base

FAKE = {
    'Description': 'Blah blah',
    'Parameters': {'key_name': {'type': 'string'}},
    'ParameterGroups': [{'label': 'Group 1', 'parameters': ['key_name']}],
}


class TestTemplate(base.TestCase):
    def test_basic(self):
        sot = template.Template()
        self.assertFalse(sot.allow_create)
        self.assertFalse(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertFalse(sot.allow_list)

    def test_make_it(self):
        sot = template.Template(**FAKE)
        self.assertEqual(FAKE['Description'], sot.description)
        self.assertEqual(FAKE['Parameters'], sot.parameters)
        self.assertEqual(FAKE['ParameterGroups'], sot.parameter_groups)

    @mock.patch.object(resource.Resource, '_translate_response')
    def test_validate(self, mock_translate):
        sess = mock.Mock()
        sot = template.Template()
        tmpl = mock.Mock()
        body = {'template': tmpl}

        sot.validate(sess, tmpl)

        sess.post.assert_called_once_with('/validate', json=body)
        mock_translate.assert_called_once_with(sess.post.return_value)

    @mock.patch.object(resource.Resource, '_translate_response')
    def test_validate_with_env(self, mock_translate):
        sess = mock.Mock()
        sot = template.Template()
        tmpl = mock.Mock()
        env = mock.Mock()
        body = {'template': tmpl, 'environment': env}

        sot.validate(sess, tmpl, environment=env)

        sess.post.assert_called_once_with('/validate', json=body)
        mock_translate.assert_called_once_with(sess.post.return_value)

    @mock.patch.object(resource.Resource, '_translate_response')
    def test_validate_with_template_url(self, mock_translate):
        sess = mock.Mock()
        sot = template.Template()
        template_url = 'http://host1'
        body = {'template': None, 'template_url': template_url}

        sot.validate(sess, None, template_url=template_url)

        sess.post.assert_called_once_with('/validate', json=body)
        mock_translate.assert_called_once_with(sess.post.return_value)

    @mock.patch.object(resource.Resource, '_translate_response')
    def test_validate_with_ignore_errors(self, mock_translate):
        sess = mock.Mock()
        sot = template.Template()
        tmpl = mock.Mock()
        body = {'template': tmpl}

        sot.validate(sess, tmpl, ignore_errors='123,456')

        sess.post.assert_called_once_with(
            '/validate?ignore_errors=123%2C456', json=body
        )
        mock_translate.assert_called_once_with(sess.post.return_value)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.container_infrastructure_management.v1 import cluster_template
from openstack import exceptions
from openstack.tests.unit import base


cluster_template_obj = dict(
    apiserver_port=12345,
    cluster_distro='fake-distro',
    coe='fake-coe',
    created_at='fake-date',
    dns_nameserver='8.8.8.8',
    docker_volume_size=1,
    external_network_id='public',
    fixed_network=None,
    flavor_id='fake-flavor',
    https_proxy=None,
    human_id=None,
    image_id='fake-image',
    insecure_registry='https://192.168.0.10',
    keypair_id='fake-key',
    labels={},
    links={},
    master_flavor_id=None,
    name='fake-cluster-template',
    network_driver='fake-driver',
    no_proxy=None,
    public=False,
    registry_enabled=False,
    server_type='vm',
    tls_disabled=False,
    updated_at=None,
    uuid='fake-uuid',
    volume_driver=None,
)


class TestClusterTemplates(base.TestCase):
    def _compare_clustertemplates(self, exp, real):
        self.assertDictEqual(
            cluster_template.ClusterTemplate(**exp).to_dict(computed=False),
            real.to_dict(computed=False),
        )

    def get_mock_url(
        self,
        service_type='container-infrastructure-management',
        base_url_append=None,
        append=None,
        resource=None,
    ):
        return super().get_mock_url(
            service_type=service_type,
            resource=resource,
            append=append,
            base_url_append=base_url_append,
        )

    def test_list_cluster_templates_without_detail(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(resource='clustertemplates'),
                    json=dict(clustertemplates=[cluster_template_obj]),
                )
            ]
        )
        cluster_templates_list = self.cloud.list_cluster_templates()
        self._compare_clustertemplates(
            cluster_template_obj,
            cluster_templates_list[0],
        )
        self.assert_calls()

    def test_list_cluster_templates_with_detail(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(resource='clustertemplates'),
                    json=dict(clustertemplates=[cluster_template_obj]),
                )
            ]
        )
        cluster_templates_list = self.cloud.list_cluster_templates(detail=True)
        self._compare_clustertemplates(
            cluster_template_obj,
            cluster_templates_list[0],
        )
        self.assert_calls()

    def test_search_cluster_templates_by_name(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(resource='clustertemplates'),
                    json=dict(clustertemplates=[cluster_template_obj]),
                )
            ]
        )

        cluster_templates = self.cloud.search_cluster_templates(
            name_or_id='fake-cluster-template'
        )

        self.assertEqual(1, len(cluster_templates))
        self.assertEqual('fake-uuid', cluster_templates[0]['uuid'])
        self.assert_calls()

    def test_search_cluster_templates_not_found(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(resource='clustertemplates'),
                    json=dict(clustertemplates=[cluster_template_obj]),
                )
            ]
        )

        cluster_templates = self.cloud.search_cluster_templates(
            name_or_id='non-existent'
        )

        self.assertEqual(0, len(cluster_templates))
        self.assert_calls()

    def test_get_cluster_template(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(resource='clustertemplates'),
                    json=dict(clustertemplates=[cluster_template_obj]),
                )
            ]
        )

        r = self.cloud.get_cluster_template('fake-cluster-template')
        self.assertIsNotNone(r)
        self._compare_clustertemplates(
            cluster_template_obj,
            r,
        )
        self.assert_calls()

    def test_get_cluster_template_not_found(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(resource='clustertemplates'),
                    json=dict(clustertemplates=[]),
                )
            ]
        )
        r = self.cloud.get_cluster_template('doesNotExist')
        self.assertIsNone(r)
        self.assert_calls()

    def test_create_cluster_template(self):
        json_response = cluster_template_obj.copy()
        kwargs = dict(
            name=cluster_template_obj['name'],
            image_id=cluster_template_obj['image_id'],
            keypair_id=cluster_template_obj['keypair_id'],
            coe=cluster_template_obj['coe'],
        )
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(resource='clustertemplates'),
                    json=json_response,
                    validate=dict(json=kwargs),
                )
            ]
        )
        response = self.cloud.create_cluster_template(**kwargs)
        self._compare_clustertemplates(json_response, response)

        self.assert_calls()

    def test_create_cluster_template_exception(self):
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(resource='clustertemplates'),
                    status_code=403,
                )
            ]
        )
        # TODO(mordred) requests here doens't give us a great story
        # for matching the old error message text. Investigate plumbing
        # an error message in to the adapter call so that we can give a
        # more informative error. Also, the test was originally catching
        # SDKException - but for some reason testtools will not
        # match the more specific HTTPError, even though it's a subclass
        # of SDKException.
        with self.assertRaises(exceptions.ForbiddenException):
            self.cloud.create_cluster_template('fake-cluster-template')
        self.assert_calls()

    def test_delete_cluster_template(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(resource='clustertemplates'),
                    json=dict(clustertemplates=[cluster_template_obj]),
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        resource='clustertemplates/fake-uuid'
                    ),
                ),
            ]
        )
        self.cloud.delete_cluster_template('fake-uuid')
        self.assert_calls()

    def test_update_cluster_template(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(resource='clustertemplates'),
                    json=dict(clustertemplates=[cluster_template_obj]),
                ),
                dict(
                    method='PATCH',
                    uri=self.get_mock_url(
                        resource='clustertemplates/fake-uuid'
                    ),
                    status_code=200,
                    validate=dict(
                        json=[
                            {
                                'op': 'replace',
                                'path': '/name',
                                'value': 'new-cluster-template',
                            }
                        ]
                    ),
                ),
            ]
        )
        new_name = 'new-cluster-template'
        updated = self.cloud.update_cluster_template(
            'fake-uuid', name=new_name
        )
        self.assertEqual(new_name, updated.name)
        self.assert_calls()

    def test_coe_get_cluster_template(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(resource='clustertemplates'),
                    json=dict(clustertemplates=[cluster_template_obj]),
                )
            ]
        )

        r = self.cloud.get_cluster_template('fake-cluster-template')
        self.assertIsNotNone(r)
        self._compare_clustertemplates(
            cluster_template_obj,
            r,
        )
        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class ResourceFilter(resource.Resource):
    """Resource Filter"""

    resources_key = "resource_filters"
    base_path = "/resource_filters"

    _query_mapping = resource.QueryParameters(
        'resource',
        include_pagination_defaults=False,
    )

    # Capabilities
    allow_list = True

    # resource_filters introduced in 3.33
    _max_microversion = '3.33'

    #: Properties
    #: The list of filters that are applicable to the specified resource.
    filters = resource.Body('filters', type=list)
    #: The resource that the filters will be applied to.
    resource = resource.Body('resource', type=str)
